#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
from datetime import date, datetime
from typing import Any, Dict, Optional

import boto3

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def json_serial(obj: Any) -> str:
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def handler(event: Dict[str, Any], context: Optional[Dict[str, Any]]) -> Any:
    logger.info(f"Lambda metadata: {json.dumps(event)} (type = {type(event)})")

    name = event.get("name", None)
    if name is None:
        raise ValueError("Invalid 'name': %s", name)

    eks = boto3.client("eks")
    return json.loads(json.dumps(eks.describe_cluster(name=name), default=json_serial))



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
import shutil
import sys
from typing import Any, cast

import aws_cdk.aws_cognito as cognito
import aws_cdk.aws_ec2 as ec2
import aws_cdk.aws_iam as iam
import aws_cdk.aws_lambda_python as lambda_python
import aws_cdk.aws_sam as sam
import aws_cdk.aws_ssm as ssm
import jsii
from aws_cdk import aws_lambda, core
from aws_cdk.core import App, Aspects, Construct, Duration, Environment, IAspect, IConstruct, Stack, Tags
from botocore.exceptions import ClientError

from aws_orbit.models.context import Context, ContextSerDe
from aws_orbit.remote_files.cdk import _lambda_path
from aws_orbit.remote_files.cdk.team_builders.codeartifact import DeployCodeArtifact
from aws_orbit.services import cognito as orbit_cognito
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)

# Client Connections
CA_CLIENT = boto3_client("codeartifact")

"""
Custom Implementation of NonPathRole is to extract EKS Roles (NodeGroup and Fargate).
https://github.com/kubernetes-sigs/aws-iam-authenticator/issues/268
"""


class NonPathRole(iam.Role):
    def __init__(self, scope: core.Construct, id: str, **kwargs: Any):
        super().__init__(scope, id, **kwargs)


class Env(Stack):
    def __init__(
        self,
        scope: Construct,
        id: str,
        context: "Context",
    ) -> None:
        self.scope = scope
        self.id = id
        self.context = context
        super().__init__(
            scope=scope,
            id=id,
            stack_name=id,
            env=Environment(account=self.context.account_id, region=self.context.region),
        )
        Tags.of(scope=cast(IConstruct, self)).add(key="Env", value=f"orbit-{self.context.name}")
        if self.context.networking.vpc_id is None:
            raise ValueError("self.context.networking.vpc_id is None.")
        if self.context.networking.availability_zones is None:
            raise ValueError("self.context.networking.availability_zones is None.")

        # Checks if CodeArtifact exists outside of the scope of Orbit, else creates it.
        if not self.context.codeartifact_domain and not self.context.codeartifact_repository:
            if not self._check_ca_domain_existence() and not self._check_ca_repo_existence():
                _logger.debug("Deploying Codeartifact resources via Orbit Environment")
                DeployCodeArtifact(self, id="CodeArtifact-from-Env")
            else:
                _logger.debug(
                    "Detected CodeArtifact domain: aws-orbit and CodeArtifact Repo: python-repository in the account"
                )
        else:
            _logger.debug(
                "Detected CodeArtifact resources from the framework's metadata, and not deploying Codeartifact"
            )

        self.i_vpc = ec2.Vpc.from_vpc_attributes(
            scope=self,
            id="vpc",
            vpc_id=self.context.networking.vpc_id,
            availability_zones=self.context.networking.availability_zones,
        )
        self.role_eks_cluster = self._create_role_cluster()
        self.role_eks_env_nodegroup = self._create_env_nodegroup_role()
        self.role_fargate_profile = self._create_role_fargate_profile()
        self.role_cluster_autoscaler = self._create_cluster_autoscaler_role()
        if self.context.user_pool_id:
            self.context.cognito_users_url = orbit_cognito.get_users_url(
                user_pool_id=self.context.user_pool_id, region=self.context.region
            )
            cognito_pool_arn: str = orbit_cognito.get_pool_arn(
                user_pool_id=self.context.user_pool_id, region=self.context.region, account=self.context.account_id
            )
            self.user_pool: cognito.UserPool = self._get_user_pool(user_pool_arn=cognito_pool_arn)
        else:
            raise Exception("Missing Cognito User Pool ID ('user_pool_id') ")
        self.user_pool_client = self._create_user_pool_client()
        self.identity_pool = self._create_identity_pool()
        self.token_validation_lambda = self._create_token_validation_lambda()
        self.eks_service_lambda = self._create_eks_service_lambda()
        self.cluster_pod_security_group = self._create_cluster_pod_security_group()
        self.context_parameter = self._create_manifest_parameter()
        self._create_post_authentication_lambda()

    def _check_ca_domain_existence(self) -> bool:
        try:
            CA_CLIENT.describe_domain(domain="aws-orbit")
        except ClientError as ex:
            if ex.response["Error"]["Code"] == "ResourceNotFoundException":
                return False
            else:
                raise ex
        return True

    def _check_ca_repo_existence(self) -> bool:
        try:
            CA_CLIENT.describe_repository(domain="aws-orbit", repository="python-repository")
        except ClientError as ex:
            if ex.response["Error"]["Code"] == "ResourceNotFoundException":
                return False
            else:
                raise ex
        return True

    def _create_role_cluster(self) -> iam.Role:
        name: str = f"orbit-{self.context.name}-{self.context.region}-eks-cluster-role"
        role = iam.Role(
            scope=self,
            id=name,
            role_name=name,
            assumed_by=cast(
                iam.IPrincipal,
                iam.CompositePrincipal(
                    iam.ServicePrincipal("eks.amazonaws.com"),
                    iam.ServicePrincipal("eks-fargate-pods.amazonaws.com"),
                ),
            ),
            managed_policies=[
                iam.ManagedPolicy.from_aws_managed_policy_name(managed_policy_name="AmazonEKSClusterPolicy"),
                iam.ManagedPolicy.from_aws_managed_policy_name(managed_policy_name="AmazonEKSServicePolicy"),
                iam.ManagedPolicy.from_aws_managed_policy_name(managed_policy_name="AmazonEKSVPCResourceController"),
            ],
            inline_policies={
                "Extras": iam.PolicyDocument(
                    statements=[
                        iam.PolicyStatement(
                            actions=[
                                "elasticloadbalancing:*",
                                "ec2:CreateSecurityGroup",
                                "ec2:Describe*",
                                "cloudwatch:PutMetricData",
                                "iam:ListAttachedRolePolicies",
                            ],
                            resources=["*"],
                        ),
                        iam.PolicyStatement(
                            effect=iam.Effect.ALLOW,
                            actions=["iam:AttachRolePolicy", "iam:PutRolePolicy", "s3:*"],
                            resources=[
                                "arn:aws:iam::*:role/aws-service-role/s3.data-source.lustre.fsx.amazonaws.com/",
                                f"{self.context.scratch_bucket_arn}",
                                f"{self.context.scratch_bucket_arn}/*",
                            ],
                        ),
                        iam.PolicyStatement(
                            effect=iam.Effect.ALLOW,
                            actions=[
                                "iam:CreateServiceLinkedRole",
                                "s3:ListBucket",
                                "fsx:CreateFileSystem",
                                "fsx:DeleteFileSystem",
                                "fsx:DescribeFileSystems",
                                "fsx:TagResource",
                            ],
                            resources=["*"],
                        ),
                    ]
                )
            },
        )

        return role

    def _create_role_fargate_profile(self) -> NonPathRole:
        name: str = f"orbit-{self.context.name}-{self.context.region}-eks-fargate-profile-role"
        role = NonPathRole(
            scope=self,
            id=name,
            role_name=name,
            assumed_by=cast(
                iam.IPrincipal,
                iam.CompositePrincipal(
                    iam.ServicePrincipal("eks.amazonaws.com"),
                    iam.ServicePrincipal("eks-fargate-pods.amazonaws.com"),
                ),
            ),
            managed_policies=[
                iam.ManagedPolicy.from_aws_managed_policy_name(
                    managed_policy_name="AmazonEC2ContainerRegistryReadOnly"
                ),
                iam.ManagedPolicy.from_aws_managed_policy_name(
                    managed_policy_name="AmazonEKSFargatePodExecutionRolePolicy"
                ),
            ],
            inline_policies={
                "Logging": iam.PolicyDocument(
                    statements=[
                        iam.PolicyStatement(
                            effect=iam.Effect.ALLOW,
                            actions=[
                                "logs:CreateLogStream",
                                "logs:CreateLogGroup",
                                "logs:DescribeLogStreams",
                                "logs:PutLogEvents",
                            ],
                            resources=["*"],
                        )
                    ]
                )
            },
        )
        return role

    def _create_env_nodegroup_role(self) -> NonPathRole:
        name: str = f"orbit-{self.context.name}-{self.context.region}-eks-nodegroup-role"
        role = NonPathRole(
            scope=self,
            id=name,
            role_name=name,
            assumed_by=cast(iam.IPrincipal, iam.ServicePrincipal(service="ec2.amazonaws.com")),
            managed_policies=[
                iam.ManagedPolicy.from_aws_managed_policy_name(managed_policy_name="AmazonEKSWorkerNodePolicy"),
                iam.ManagedPolicy.from_aws_managed_policy_name(managed_policy_name="AmazonEKS_CNI_Policy"),
                iam.ManagedPolicy.from_aws_managed_policy_name(
                    managed_policy_name="AmazonEC2ContainerRegistryReadOnly"
                ),
                iam.ManagedPolicy.from_aws_managed_policy_name(managed_policy_name="AmazonSSMManagedInstanceCore"),
            ],
        )
        return role

    def _create_cluster_autoscaler_role(self) -> iam.Role:
        name: str = f"orbit-{self.context.name}-{self.context.region}-cluster-autoscaler-role"
        return iam.Role(
            scope=self,
            id=name,
            role_name=name,
            assumed_by=cast(iam.IPrincipal, iam.ServicePrincipal("eks.amazonaws.com")),
            inline_policies={
                "Logging": iam.PolicyDocument(
                    statements=[
                        iam.PolicyStatement(
                            effect=iam.Effect.ALLOW,
                            actions=[
                                "autoscaling:DescribeAutoScalingGroups",
                                "autoscaling:DescribeAutoScalingInstances",
                                "autoscaling:DescribeLaunchConfigurations",
                                "autoscaling:DescribeTags",
                                "autoscaling:SetDesiredCapacity",
                                "autoscaling:TerminateInstanceInAutoScalingGroup",
                                "ec2:DescribeLaunchTemplateVersions",
                            ],
                            resources=["*"],
                        )
                    ]
                )
            },
        )

    def _get_user_pool(self, user_pool_arn: str) -> cognito.UserPool:
        return cast(
            cognito.UserPool,
            cognito.UserPool.from_user_pool_arn(scope=self, id="orbit-user-pool", user_pool_arn=user_pool_arn),
        )

    def _create_user_pool_client(self) -> cognito.UserPoolClient:
        return cognito.UserPoolClient(
            scope=self,
            id="user-pool-client",
            user_pool=cast(cognito.IUserPool, self.user_pool),
            auth_flows=cognito.AuthFlow(user_srp=True, admin_user_password=False, custom=False),
            generate_secret=True,
            prevent_user_existence_errors=True,
            user_pool_client_name="orbit",
        )

    def _create_identity_pool(self) -> cognito.CfnIdentityPool:
        provider_name = (
            self.user_pool.user_pool_provider_name
            if hasattr(self.user_pool, "user_pool_provider_name")
            else f"cognito-idp.{self.context.region}.amazonaws.com/{self.context.user_pool_id}"
        )

        pool = cognito.CfnIdentityPool(
            scope=self,
            id="identity-pool",
            identity_pool_name=self.id.replace("-", "_"),
            allow_unauthenticated_identities=False,
            allow_classic_flow=False,
            cognito_identity_providers=[
                cognito.CfnIdentityPool.CognitoIdentityProviderProperty(
                    provider_name=provider_name,
                    client_id=self.user_pool_client.user_pool_client_id,
                )
            ],
        )
        name = f"{self.id}-{self.context.region}-cognito-auth-identity-role"
        authenticated_role = iam.Role(
            scope=self,
            id=name,
            role_name=name,
            assumed_by=cast(
                iam.IPrincipal,
                iam.FederatedPrincipal(
                    federated="cognito-identity.amazonaws.com",
                    conditions={
                        "StringEquals": {"cognito-identity.amazonaws.com:aud": pool.ref},
                        "ForAnyValue:StringLike": {"cognito-identity.amazonaws.com:amr": "authenticated"},
                    },
                    assume_role_action="sts:AssumeRoleWithWebIdentity",
                ),
            ),
            inline_policies={
                "cognito-default": iam.PolicyDocument(
                    statements=[
                        iam.PolicyStatement(
                            actions=[
                                "mobileanalytics:PutEvents",
                                "cognito-sync:*",
                                "cognito-identity:*",
                            ],
                            resources=["*"],
                        )
                    ]
                ),
                "team-context-parameter": iam.PolicyDocument(
                    statements=[
                        iam.PolicyStatement(
                            actions=["ssm:DescribeParameters", "ssm:GetParameters"],
                            resources=[
                                f"arn:aws:ssm:{self.context.region}:{self.context.account_id}:"
                                f"parameter/orbit/{self.context.name}/teams/*"
                            ],
                        )
                    ]
                ),
            },
        )
        name = f"{self.id}-{self.context.region}-cognito-unauth-identity-role"
        unauthenticated_role = iam.Role(
            scope=self,
            id=name,
            role_name=name,
            assumed_by=cast(
                iam.IPrincipal,
                iam.FederatedPrincipal(
                    federated="cognito-identity.amazonaws.com",
                    conditions={
                        "StringEquals": {"cognito-identity.amazonaws.com:aud": pool.ref},
                        "ForAnyValue:StringLike": {"cognito-identity.amazonaws.com:amr": "unauthenticated"},
                    },
                    assume_role_action="sts:AssumeRoleWithWebIdentity",
                ),
            ),
            inline_policies={
                "cognito-default": iam.PolicyDocument(
                    statements=[
                        iam.PolicyStatement(
                            actions=[
                                "mobileanalytics:PutEvents",
                                "cognito-sync:*",
                            ],
                            resources=["*"],
                        )
                    ]
                )
            },
        )
        cognito.CfnIdentityPoolRoleAttachment(
            scope=self,
            id=f"{self.id}-role-attachment",
            identity_pool_id=pool.ref,
            roles={
                "authenticated": authenticated_role.role_arn,
                "unauthenticated": unauthenticated_role.role_arn,
            },
        )
        return pool

    def _create_token_validation_lambda(self) -> aws_lambda.Function:

        return lambda_python.PythonFunction(
            scope=self,
            id="token_validation_lambda",
            function_name=f"orbit-{self.context.name}-token-validation",
            entry=_lambda_path("token_validation"),
            index="index.py",
            handler="handler",
            runtime=aws_lambda.Runtime.PYTHON_3_8,
            timeout=Duration.seconds(5),
            environment={
                "COGNITO_USER_POOL_ID": self.user_pool.user_pool_id,
                "REGION": self.context.region,
                "COGNITO_USER_POOL_CLIENT_ID": self.user_pool_client.user_pool_client_id,
            },
            initial_policy=[
                iam.PolicyStatement(
                    effect=iam.Effect.ALLOW,
                    actions=["ec2:Describe*", "logs:Create*", "logs:PutLogEvents", "logs:Describe*"],
                    resources=["*"],
                )
            ],
        )

    def _create_eks_service_lambda(self) -> aws_lambda.Function:

        return lambda_python.PythonFunction(
            scope=self,
            id="eks_service_lambda",
            function_name=f"orbit-{self.context.name}-eks-service-handler",
            entry=_lambda_path("eks_service_handler"),
            index="index.py",
            handler="handler",
            runtime=aws_lambda.Runtime.PYTHON_3_8,
            timeout=Duration.seconds(5),
            environment={
                "REGION": self.context.region,
            },
            initial_policy=[
                iam.PolicyStatement(
                    effect=iam.Effect.ALLOW,
                    actions=["eks:List*", "eks:Describe*"],
                    resources=[
                        f"arn:aws:eks:{self.context.region}:{self.context.account_id}:cluster/orbit-*",
                        f"arn:aws:eks:{self.context.region}:{self.context.account_id}:nodegroup/orbit-*/*/*",
                    ],
                )
            ],
        )

    def _create_post_authentication_lambda(self) -> None:
        k8s_layer_name = f"orbit-{self.context.name}-k8s-base-layer"

        sam_app = sam.CfnApplication(
            scope=self,
            id="awscli_kubectl_helm_lambda_layer_sam",
            location=sam.CfnApplication.ApplicationLocationProperty(
                application_id="arn:aws:serverlessrepo:us-east-1:903779448426:applications/lambda-layer-kubectl",
                semantic_version="2.0.0",
            ),
            parameters={"LayerName": k8s_layer_name},
        )

        role_arn = cast(str, self.context.toolkit.admin_role_arn)

        lambda_python.PythonFunction(
            scope=self,
            id="cognito_post_authentication_lambda",
            function_name=f"orbit-{self.context.name}-post-authentication",
            entry=_lambda_path("cognito_post_authentication"),
            index="index.py",
            handler="handler",
            runtime=aws_lambda.Runtime.PYTHON_3_7,
            timeout=Duration.seconds(300),
            role=iam.Role.from_role_arn(scope=self, id="cognito-post-auth-role", role_arn=role_arn),
            environment={
                "REGION": self.context.region,
                "ORBIT_ENV": self.context.name,
                "ACCOUNT_ID": self.context.account_id,
            },
            memory_size=128,
        ).add_permission(
            id="cognito_post_auth_resource_policy",
            principal=cast(iam.IPrincipal, iam.ServicePrincipal("cognito-idp.amazonaws.com")),
            action="lambda:InvokeFunction",
            source_arn=(
                f"arn:aws:cognito-idp:{self.context.region}:{self.context.account_id}:"
                f"userpool/{self.user_pool.user_pool_id}"
            ),
        )

        lambda_python.PythonFunction(
            scope=self,
            id="cognito_post_authentication_k8s_lambda",
            entry=_lambda_path("cognito_post_authentication"),
            function_name=f"orbit-{self.context.name}-post-auth-k8s-manage",
            index="k8s_manage.py",
            handler="handler",
            runtime=aws_lambda.Runtime.PYTHON_3_7,
            timeout=Duration.seconds(300),
            role=iam.Role.from_role_arn(scope=self, id="cognito-post-auth-k8s-role", role_arn=role_arn),
            environment={
                "REGION": self.context.region,
                "PATH": "/var/lang/bin:/usr/local/bin:/usr/bin/:/bin:/opt/bin:/opt/awscli:/opt/kubectl:/opt/helm",
                "ORBIT_ENV": self.context.name,
                "ACCOUNT_ID": self.context.account_id,
                "ROLE_PREFIX": f"/{self.context.role_prefix}/" if self.context.role_prefix else "/",
                "ORBIT_API_VERSION": "v1",
                "ORBIT_API_GROUP": "orbit.aws",
            },
            layers=[
                aws_lambda.LayerVersion.from_layer_version_arn(
                    scope=self,
                    id="K8sLambdaLayer",
                    layer_version_arn=(sam_app.get_att("Outputs.LayerVersionArn").to_string()),
                )
            ],
            memory_size=256,
        )

    def _create_manifest_parameter(self) -> ssm.StringParameter:
        parameter: ssm.StringParameter = ssm.StringParameter(
            scope=self,
            id="/orbit/EnvParams",
            string_value=json.dumps(
                {
                    "EksClusterRoleArn": self.role_eks_cluster.role_arn,
                    "EksFargateProfileRoleArn": self.role_fargate_profile.role_arn,
                    "EksEnvNodegroupRoleArn": self.role_eks_env_nodegroup.role_arn,
                    "EksClusterAutoscalerRoleArn": self.role_cluster_autoscaler.role_arn,
                    "UserPoolId": self.user_pool.user_pool_id,
                    "UserPoolClientId": self.user_pool_client.user_pool_client_id,
                    "IdentityPoolId": self.identity_pool.ref,
                    "ClusterPodSecurityGroupId": self.cluster_pod_security_group.security_group_id,
                }
            ),
            type=ssm.ParameterType.STRING,
            description="Orbit Workbench Remote Env.",
            parameter_name=self.context.env_ssm_parameter_name,
            simple_name=False,
            tier=ssm.ParameterTier.INTELLIGENT_TIERING,
        )
        return parameter

    def _create_cluster_pod_security_group(self) -> ec2.SecurityGroup:
        name = f"orbit-{self.context.name}-cluster-pod-sg"
        sg = ec2.SecurityGroup(
            scope=self,
            id="cluster-pod-security-group",
            security_group_name=name,
            vpc=self.i_vpc,
        )
        Tags.of(scope=cast(IConstruct, sg)).add(key="Name", value=name)
        return sg


def main() -> None:
    _logger.debug("sys.argv: %s", sys.argv)
    if len(sys.argv) == 2:
        context: "Context" = ContextSerDe.load_context_from_ssm(env_name=sys.argv[1], type=Context)
    else:
        raise ValueError(f"Unexpected number of values in sys.argv ({len(sys.argv)}), {sys.argv}")

    outdir = os.path.join(".orbit.out", context.name, "cdk", context.env_stack_name)
    os.makedirs(outdir, exist_ok=True)
    shutil.rmtree(outdir)

    app = App(outdir=outdir)

    @jsii.implements(core.IAspect)
    class AddDeployPathIAM:
        """ Implementing CDK Aspects to add optional IAM Role prefix to IAM roles """

        def visit(self, obj: core.IConstruct) -> None:
            """ Function to implement a path pattern """
            if isinstance(obj, NonPathRole):
                cfn_role = obj.node.find_child("Resource")
                path = "/"
                cfn_role.add_property_override("Path", path)
            elif isinstance(obj, iam.Role):
                cfn_role = obj.node.find_child("Resource")
                path = f"/{context.role_prefix}/" if context.role_prefix else "/"
                cfn_role.add_property_override("Path", path)

    env_stack = Env(
        scope=app,
        id=context.env_stack_name,
        context=context,
    )
    Aspects.of(scope=cast(IConstruct, env_stack)).add(cast(IAspect, AddDeployPathIAM()))
    app.synth(force=True)


if __name__ == "__main__":
    main()



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import cast

from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def modify_vpc_endpoint(vpc_id: str, service_name: str, private_dns_enabled: bool) -> None:
    ec2_client = boto3_client("ec2")
    paginator = ec2_client.get_paginator("describe_vpc_endpoints")
    _logger.debug("Modifying VPC Endpoints for VPC: %s", vpc_id)
    response_iterator = paginator.paginate(Filters=[{"Name": "vpc-id", "Values": [vpc_id]}], MaxResults=50)
    for response in response_iterator:
        for ep in response["VpcEndpoints"]:
            ep_s_name = ep["ServiceName"]
            _logger.debug(f"ep_s_name={ep_s_name}")
            if service_name in ep["ServiceName"]:
                ec2_client.modify_vpc_endpoint(VpcEndpointId=ep["VpcEndpointId"], PrivateDnsEnabled=private_dns_enabled)


def get_env_vpc_id(env_name: str) -> str:
    ec2_client = boto3_client("ec2")
    paginator = ec2_client.get_paginator("describe_vpcs")
    response_iterator = paginator.paginate(Filters=[{"Name": "tag:Env", "Values": [f"orbit-{env_name}"]}])
    for response in response_iterator:
        for vpc in response["Vpcs"]:
            return cast(str, vpc["VpcId"])
    raise ValueError(f"VPC not found for env {env_name}.")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import TypeVar

from aws_orbit import cdk
from aws_orbit.models.context import Context, FoundationContext
from aws_orbit.services import cfn, s3

_logger: logging.Logger = logging.getLogger(__name__)


T = TypeVar("T")


def deploy(context: T) -> None:
    if not (isinstance(context, Context) or isinstance(context, FoundationContext)):
        raise ValueError("Unknown 'context' Type")
    _logger.debug("Deploying %s CDK Toolkit...", context.cdk_toolkit.stack_name)
    cdk.deploy_toolkit(context=context)


def destroy(context: T) -> None:
    if not (isinstance(context, Context) or isinstance(context, FoundationContext)):
        raise ValueError("Unknown 'context' Type")
    _logger.debug("Destroying %s CDK Toolkit...", context.cdk_toolkit.stack_name)
    if context.cdk_toolkit.s3_bucket:
        if cfn.does_stack_exist(stack_name=context.cdk_toolkit.stack_name):
            try:
                s3.delete_bucket(bucket=context.cdk_toolkit.s3_bucket)
            except Exception as ex:
                _logger.debug("Skipping Toolkit bucket deletion. Cause: %s", ex)
            cfn.destroy_stack(stack_name=context.cdk_toolkit.stack_name)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
from typing import List, Optional

from aws_orbit import bundle, remote
from aws_orbit.messages import MessagesContext, stylize
from aws_orbit.models.context import Context, ContextSerDe
from aws_orbit.services import cfn, codebuild

_logger: logging.Logger = logging.getLogger(__name__)


def build_image(
    env: str,
    dir: Optional[str],
    name: str,
    script: Optional[str],
    build_args: Optional[List[str]],
    timeout: int = 30,
    debug: bool = False,
    source_registry: Optional[str] = None,
    source_repository: Optional[str] = None,
    source_version: Optional[str] = None,
) -> None:
    with MessagesContext("Deploying Docker Image", debug=debug) as msg_ctx:
        context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env, type=Context)
        msg_ctx.info("Manifest loaded")
        if cfn.does_stack_exist(stack_name=f"orbit-{context.name}") is False:
            msg_ctx.error("Please, deploy your environment before deploying any additional docker image")
            return
        msg_ctx.progress(3)
        if dir:
            dirs = [(dir, name)]
        else:
            dirs = []
        bundle_path = bundle.generate_bundle(command_name=f"deploy_image-{name}", context=context, dirs=dirs)
        msg_ctx.progress(5)

        script_str = "NO_SCRIPT" if script is None else script
        source_str = "NO_REPO" if source_registry is None else f"{source_registry} {source_repository} {source_version}"
        build_args = [] if build_args is None else build_args
        buildspec = codebuild.generate_spec(
            context=context,
            plugins=False,
            cmds_build=[
                f"orbit remote --command build_image " f"{env} {name} {script_str} {source_str} {' '.join(build_args)}"
            ],
            changeset=None,
        )
        msg_ctx.progress(6)

        remote.run(
            command_name=f"deploy_image-{name}",
            context=context,
            bundle_path=bundle_path,
            buildspec=buildspec,
            codebuild_log_callback=msg_ctx.progress_bar_callback,
            timeout=timeout,
        )
        msg_ctx.info("Docker Image deploy into ECR")

        address = (
            f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com/orbit-{context.name}/{name}"
            if name in [n.replace("_", "-") for n in context.images.names]
            else f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com/orbit-{context.name}/users/{name}"
        )

        msg_ctx.info(f"ECR Image Address={address}")
        msg_ctx.tip(f"ECR Image Address: {stylize(address, underline=True)}")
        msg_ctx.progress(100)


def build_podsetting(
    env_name: str,
    team_name: str,
    podsetting: str,
    debug: bool = False,
) -> None:
    # Do some quick validation and fail fast
    with MessagesContext("Creating PodSetting", debug=debug) as msg_ctx:
        ps = json.loads(podsetting)
        msg_ctx.progress(3)
        _logger.debug(ps)
        if not ps["description"] or not ps["name"]:
            print("raise an error...")
            msg_ctx.error("Please, make sure the fields description and name are present")
            return
        ps_name = ps["name"]
        msg_ctx.progress(10)
        resources = ps.get("resources", None)
        if resources is not None:
            if (
                "nvidia.com/gpu" in resources["limits"]
                or "nvidia.com/gpu" in resources["requests"]
                or "node.kubernetes.io/instance-type" in resources["requests"]
                or "node.kubernetes.io/instance-type" in resources["limits"]
            ):
                if ps.get("node-group") is None and ps.get("instance-type") is None:
                    msg_ctx.error(
                        "If you request a GPU, please define a node-group or instance-type to match your cluster / GPU"
                    )
                    return
        msg_ctx.info("JSON payload validated")
        msg_ctx.progress(15)
        # Looks ok...call the SDK to build and deploy
        try:
            import aws_orbit_sdk.controller as controller

            controller.build_podsetting(env_name=env_name, team_name=team_name, podsetting=podsetting, debug=debug)
        except ImportError:
            raise ImportError("Make sure the Orbit SDK is installed")

        msg_ctx.info(f"PodSetting {ps_name} created")
        msg_ctx.progress(100)



import logging
import time
from typing import Any, Dict, Iterator, List

from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def _filter_env_filesystems(page: Dict[str, Any], env_name: str) -> Iterator[str]:
    for fs in page["FileSystems"]:
        for tag in fs["Tags"]:
            if tag["Key"] == "Env" and tag["Value"] == f"orbit-{env_name}":
                yield fs["FileSystemId"]


def _filter_filesystems_by_tag(page: Dict[str, Any], key: str, value: str) -> Iterator[str]:
    for fs in page["FileSystems"]:
        for tag in fs["Tags"]:
            if tag["Key"] == key and tag["Value"] == value:
                yield fs["FileSystemId"]


def fetch_env_filesystems(env_name: str) -> Iterator[str]:
    client = boto3_client("efs")
    paginator = client.get_paginator("describe_file_systems")
    for page in paginator.paginate():
        for fs_id in _filter_env_filesystems(page=page, env_name=env_name):
            yield fs_id


def fetch_filesystems_by_tag(key: str, value: str) -> Iterator[str]:
    client = boto3_client("efs")
    paginator = client.get_paginator("describe_file_systems")
    for page in paginator.paginate():
        for fs_id in _filter_filesystems_by_tag(page=page, key=key, value=value):
            yield fs_id


def fetch_targets(fs_id: str) -> Iterator[str]:
    client = boto3_client("efs")
    paginator = client.get_paginator("describe_mount_targets")
    for page in paginator.paginate(FileSystemId=fs_id):
        for target in page["MountTargets"]:
            yield target["MountTargetId"]


def delete_targets(fs_id: str) -> None:
    client = boto3_client("efs")
    for target in fetch_targets(fs_id=fs_id):
        try:
            _logger.debug("Deleting fs target %s (%s)", target, fs_id)
            client.delete_mount_target(MountTargetId=target)
        except client.exceptions.MountTargetNotFound:
            _logger.warning("Ignoring MountTargetId %s deletion cause it does not exist anymore.", target)
    while client.describe_file_systems(FileSystemId=fs_id)["FileSystems"][0]["NumberOfMountTargets"] > 0:
        _logger.debug("Waiting FileSystemId %s be empty...", fs_id)
        time.sleep(3)


def delete_env_filesystems(env_name: str) -> List[str]:
    client = boto3_client("efs")
    fs_ids: List[str] = []
    for fs_id in fetch_env_filesystems(env_name=env_name):
        fs_ids.append(fs_id)
        delete_targets(fs_id=fs_id)
        _logger.debug("Deleting fs %s", fs_id)
        client.delete_file_system(FileSystemId=fs_id)
    return fs_ids


def delete_filesystems_by_tag(key: str, value: str) -> List[str]:
    client = boto3_client("efs")
    fs_ids: List[str] = []
    for fs_id in fetch_filesystems_by_tag(key=key, value=value):
        fs_ids.append(fs_id)
        delete_targets(fs_id=fs_id)
        client.delete_file_system(FileSystemId=fs_id)
    return fs_ids


def delete_filesystems_by_team(env_name: str, team_name: str) -> None:
    fs_name: str = f"orbit-{env_name}-{team_name}-fs"
    _logger.debug("fs_name: %s...", fs_name)
    delete_filesystems_by_tag(key="Name", value=fs_name)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
from typing import List, Optional, Tuple, cast

from aws_orbit import docker, sh
from aws_orbit.models.context import Context, ContextSerDe
from aws_orbit.models.manifest import ImageManifest
from aws_orbit.services import ecr

_logger: logging.Logger = logging.getLogger(__name__)


def build_image(args: Tuple[str, ...]) -> None:
    if len(args) < 4:
        raise ValueError("Unexpected number of values in args.")
    env: str = args[0]
    image_name: str = args[1]
    script: Optional[str] = args[2] if args[2] != "NO_SCRIPT" else None
    source_registry: Optional[str]
    if args[3] != "NO_REPO":
        if len(args) < 6:
            raise Exception("Source registry is defined without 'source_repository' or 'source_version' ")
        source_registry = args[3]
        source_repository: str = args[4]
        source_version: str = args[5]
        build_args = args[6:]
        _logger.info("replicating image %s: %s %s:%s", image_name, source_registry, source_repository, source_version)
    else:
        _logger.info("building image %s: %s", image_name, script)
        build_args = args[4:]
        source_registry = None
    _logger.debug("args: %s", args)
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env, type=Context)

    docker.login(context=context)
    _logger.debug("DockerHub and ECR Logged in")

    ecr_repo = (
        f"orbit-{context.name}/{image_name}"
        if image_name in [n.replace("_", "-") for n in context.images.names]
        else f"orbit-{context.name}/users/{image_name}"
    )
    if not ecr.describe_repositories(repository_names=[ecr_repo]):
        ecr.create_repository(repository_name=ecr_repo, env_name=context.name)

    image_def: Optional["ImageManifest"] = getattr(context.images, image_name.replace("-", "_"), None)
    _logger.debug("image def: %s", image_def)

    if source_registry:
        docker.replicate_image(
            context=context,
            image_name=image_name,
            deployed_name=ecr_repo,
            source=source_registry,
            source_repository=source_repository,
            source_version=source_version,
        )
    elif image_def is None or image_def.get_source(account_id=context.account_id, region=context.region) == "code":
        path = os.path.join(os.getcwd(), image_name)
        if not os.path.exists(path):
            bundle_dir = os.path.join(os.getcwd(), "bundle", image_name)
            if os.path.exists(bundle_dir):
                path = bundle_dir
            else:
                raise RuntimeError(f"Unable to locate source in {path} or {bundle_dir}")

        _logger.debug("path: %s", path)
        if script is not None:
            sh.run(f"sh {script}", cwd=path)
        tag = cast(str, image_def.version if image_def else "latest")
        docker.deploy_image_from_source(
            context=context, dir=path, name=ecr_repo, tag=tag, build_args=cast(Optional[List[str]], build_args)
        )
    else:
        docker.replicate_image(context=context, image_name=image_name, deployed_name=ecr_repo)
    _logger.debug("Docker Image Deployed to ECR")



import logging
from typing import TYPE_CHECKING, List

from kubernetes import config

_logger: logging.Logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from aws_orbit.models.context import Context


def get_k8s_context(context: "Context") -> str:
    try:
        contexts: List[str] = [str(c["name"]) for c in config.list_kube_config_contexts()[0]]
    except config.config_exception.ConfigException as e:
        _logger.exception("Context not found")
        raise e
    expected_domain: str = f"@orbit-{context.name}.{context.region}.eksctl.io"
    for ctx in contexts:
        if ctx.endswith(expected_domain):
            return ctx
    raise RuntimeError(f"Context not found for domain: {expected_domain}")



import json
import logging
from typing import Any, Dict, cast

from botocore.exceptions import ClientError

from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def get_secret_value(secret_id: str) -> Dict[str, Any]:
    client = boto3_client("secretsmanager")

    try:
        _logger.debug("Getting Secret: %s", secret_id)
        get_secret_value_response = client.get_secret_value(SecretId=secret_id)
    except ClientError as e:
        _logger.exception(e)
        return {}
    else:
        return cast(Dict[str, Any], json.loads(get_secret_value_response.get("SecretString", "{}")))


def put_secret_value(secret_id: str, secret: Dict[str, Any]) -> None:
    client = boto3_client("secretsmanager")

    try:
        _logger.debug("Creating Secret: %s", secret_id)
        client.create_secret(Name=secret_id, SecretString="{}")
    except ClientError as e:
        if e.response["Error"]["Code"] == "ResourceExistsException":
            _logger.info("Secret %s exists, ignoring", secret_id)
        else:
            _logger.exception(e)
            raise e

    try:
        _logger.debug("Putting Secret Value: %s", secret_id)
        client.put_secret_value(SecretId=secret_id, SecretString=json.dumps(secret))
    except ClientError as e:
        _logger.exception(e)


def delete_docker_credentials(secret_id: str) -> None:
    client = boto3_client("secretsmanager")

    try:
        _logger.debug("Deleting Secret: %s", secret_id)
        client.delete_secret(SecretId=secret_id, ForceDeleteWithoutRecovery=True)
    except ClientError as e:
        _logger.exception(e)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
from typing import Any, Dict, Optional

import aws_orbit_sdk.controller as controller

_logger: logging.Logger = logging.getLogger(__name__)


def _set_environ(env: str, team: str, user: str) -> None:
    os.environ["AWS_ORBIT_ENV"] = env
    os.environ["AWS_ORBIT_TEAM_SPACE"] = team
    os.environ["USERNAME"] = user
    os.environ["AWS_ORBIT_S3_BUCKET"] = "Unknown"


def _wait(tasks: Dict[str, Any], delay: Optional[int], max_attempts: Optional[int], tail_logs: bool) -> bool:
    params = {"tasks": [tasks], "tail_log": tail_logs}
    if delay:
        params["delay"] = delay
    if max_attempts:
        params["maxAttempts"] = max_attempts
    no_error: bool = controller.wait_for_tasks_to_complete(**params)
    return no_error


def run_python_container(
    env: str,
    team: str,
    user: str,
    tasks: Dict[str, Any],
    wait: bool,
    delay: Optional[int],
    max_attempts: Optional[int],
    tail_logs: bool,
    debug: bool,
) -> bool:
    if debug:
        controller._logger.setLevel(logging.DEBUG)
    _set_environ(env, team, user)
    response = controller.run_python(tasks)
    if wait:
        return _wait(response, delay, max_attempts, tail_logs)
    else:
        print(json.dumps(response))
        return True


def run_notebook_container(
    env: str,
    team: str,
    user: str,
    tasks: Dict[str, Any],
    wait: bool,
    delay: Optional[int],
    max_attempts: Optional[int],
    tail_logs: bool,
    debug: bool,
) -> bool:
    if debug:
        controller._logger.setLevel(logging.DEBUG)
    _set_environ(env, team, user)
    response = controller.run_notebooks(tasks)
    if wait:
        return _wait(response, delay, max_attempts, tail_logs)
    else:
        print(json.dumps(response))
        return True



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import shutil
from typing import Optional

from aws_orbit import ORBIT_CLI_ROOT, utils
from aws_orbit.messages import MessagesContext, stylize

_logger: logging.Logger = logging.getLogger(__name__)


def write_resolve_parameters(
    manifest_name: str,
    name: str,
    filename: str,
    region: Optional[str],
) -> None:
    region_str: str = region if region is not None else utils.get_region()
    input = os.path.join(ORBIT_CLI_ROOT, "data", "init", manifest_name)
    with open(input, "r") as file:
        content: str = file.read()
    content = utils.resolve_parameters(content, dict(region=region_str, name=name))

    with open(filename, "w") as file:
        file.write(content)


def init(name: str, region: Optional[str], foundation: bool, debug: bool) -> None:
    conf_dir = "conf"
    with MessagesContext("Initializing", debug=debug) as ctx:
        conf_dir_src = os.path.join(ORBIT_CLI_ROOT, "data", "init")
        if os.path.exists(conf_dir):
            shutil.rmtree(conf_dir)
        foundation_manifest = "default-foundation.yaml"
        env_manifest = "default-env-manifest.yaml"
        shutil.copytree(src=conf_dir_src, dst=conf_dir)
        ctx.progress(50)
        name = name.lower()

        write_resolve_parameters(
            name=name,
            filename=os.path.join(conf_dir, foundation_manifest),
            region=region,
            manifest_name=foundation_manifest,
        )
        write_resolve_parameters(
            name=name, filename=os.path.join(conf_dir, env_manifest), region=region, manifest_name=env_manifest
        )
        ctx.info("Env Manifest generated into conf folder")

        ctx.progress(100)
        if foundation:
            ctx.tip(f"Recommended next step: {stylize(f'orbit deploy foundation -f {foundation_manifest}')}")

        ctx.tip(
            f"Then, fill up the manifest file ({env_manifest}) "
            f"and run: "
            f"{stylize(f'orbit env -f {env_manifest}')}"
        )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import TYPE_CHECKING, List, cast

import aws_cdk.aws_ec2 as ec2
import aws_cdk.core as core

if TYPE_CHECKING:
    from aws_orbit.models.context import Context, SubnetContext, SubnetKind


class Ec2Builder:
    @staticmethod
    def build_subnets(scope: core.Construct, subnet_manifests: List["SubnetContext"]) -> List[ec2.ISubnet]:
        return [
            ec2.PrivateSubnet.from_subnet_attributes(
                scope=scope,
                id=s.subnet_id,
                subnet_id=s.subnet_id,
                availability_zone=s.availability_zone,
                route_table_id=s.route_table_id,
            )
            for s in subnet_manifests
        ]

    @staticmethod
    def build_efs_security_group(
        scope: core.Construct, context: "Context", team_name: str, vpc: ec2.IVpc, subnet_kind: "SubnetKind"
    ) -> ec2.SecurityGroup:
        name: str = f"orbit-{context.name}-{team_name}-efs-sg"
        sg = ec2.SecurityGroup(
            scope=scope,
            id=name,
            security_group_name=name,
            vpc=vpc,
            allow_all_outbound=True,
        )
        subnets = (
            context.networking.private_subnets + context.networking.public_subnets + context.networking.isolated_subnets
        )
        for subnet in subnets:
            if subnet.kind is subnet_kind:
                sg.add_ingress_rule(
                    peer=ec2.Peer.ipv4(cast(str, subnet.cidr_block)),
                    connection=ec2.Port.tcp(port=2049),
                    description=f"Allowing internal access from subnet {subnet.subnet_id}.",
                )
        core.Tags.of(scope=cast(core.IConstruct, sg)).add(key="Name", value=name)
        return sg

    @staticmethod
    def build_team_security_group(
        scope: core.Construct, context: "Context", team_name: str, vpc: ec2.IVpc
    ) -> ec2.SecurityGroup:
        name: str = f"orbit-{context.name}-{team_name}-sg"
        sg = ec2.SecurityGroup(scope=scope, id=name, security_group_name=name, vpc=vpc, allow_all_outbound=True)
        sg.add_ingress_rule(
            peer=cast(ec2.IPeer, sg),
            connection=ec2.Port.all_traffic(),
            description=f"Allow ingress from all resources utilizing the Team ({team_name}) security group",
        )
        core.Tags.of(scope=cast(core.IConstruct, sg)).add(key="Name", value=name)
        return sg



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import time
from typing import Tuple

from aws_orbit import cleanup, plugins, sh
from aws_orbit.exceptions import FailedShellCommand
from aws_orbit.models.context import Context, ContextSerDe, FoundationContext
from aws_orbit.remote_files import cdk_toolkit, eksctl, env, foundation, helm, kubectl, teams
from aws_orbit.services import ecr, secretsmanager, ssm

_logger: logging.Logger = logging.getLogger(__name__)


def delete_image(args: Tuple[str, ...]) -> None:
    _logger.debug("args %s", args)
    env_name: str = args[0]
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=Context)

    if len(args) == 2:
        image_name: str = args[1]
    else:
        raise ValueError("Unexpected number of values in args.")

    env.deploy(context=context, eks_system_masters_roles_changes=None)
    _logger.debug("Env changes deployed")
    ecr.delete_repo(repo=f"orbit-{context.name}/{image_name}")
    _logger.debug("Docker Image Destroyed from ECR")


def destroy_team_user_resources(team_name: str) -> None:
    try:
        sh.run(
            f"bash -c 'for ns in $(kubectl get namespaces --output=jsonpath={{.items..metadata.name}} "
            f"-l orbit/team={team_name},orbit/space=team); "
            f"do kubectl delete teamspace $ns -n $ns  --force; "
            f"done'"
        )
    except FailedShellCommand:
        _logger.error("Failed toexecute command to delete teamspace object: %s")


def destroy_teams(args: Tuple[str, ...]) -> None:
    _logger.debug("args %s", args)
    env_name: str = args[0]
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=Context)
    _logger.debug("context.name %s", context.name)
    plugins.PLUGINS_REGISTRIES.load_plugins(context=context, plugin_changesets=[], teams_changeset=None)
    kubectl.write_kubeconfig(context=context)
    for team_context in context.teams:
        destroy_team_user_resources(team_context.name)
    time.sleep(60)
    _logger.debug("Plugins loaded")
    for team_context in context.teams:
        plugins.PLUGINS_REGISTRIES.destroy_team_plugins(context=context, team_context=team_context)
    _logger.debug("Plugins destroyed")
    for team_context in context.teams:
        helm.destroy_team(context=context, team_context=team_context)
    _logger.debug("Helm Charts uninstalled")
    kubectl.destroy_teams(context=context)
    _logger.debug("Kubernetes Team components destroyed")
    eksctl.destroy_teams(context=context)
    _logger.debug("EKS Team Stacks destroyed")
    teams.destroy_all(context=context)
    _logger.debug("Teams Stacks destroyed")
    ssm.cleanup_teams(env_name=context.name)


def destroy_env(args: Tuple[str, ...]) -> None:
    _logger.debug("args %s", args)
    env_name: str = args[0]
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=Context)
    _logger.debug("context.name %s", context.name)

    # Helps save time on target group issues with vpc
    cleanup.delete_istio_ingress(context=context)
    cleanup.delete_target_group(cluster_name=context.env_stack_name)

    # Delete the optional Fargate Profile
    cleanup.delete_system_fargate_profile(context=context)

    # Delete istio-system pod disruption budget; Causes a dead lock
    cleanup.delete_istio_pod_disruption_budget(context=context)

    # helm.destroy_env(context=context)
    # _logger.debug("Helm Charts uninstalled")

    # kubeflow.destroy_kubeflow(context=context)
    # _logger.debug("Kubeflow uninstalled")

    # kubectl.destroy_env(context=context)
    # _logger.debug("Kubernetes Environment components destroyed")

    eksctl.destroy_env(context=context)
    _logger.debug("EKS Environment Stacks destroyed")

    cleanup.delete_elb_security_group(cluster_name=context.env_stack_name)

    env.destroy(context=context)
    _logger.debug("Env Stack destroyed")
    cdk_toolkit.destroy(context=context)
    _logger.debug("CDK Toolkit Stack destroyed")


def destroy_foundation(args: Tuple[str, ...]) -> None:
    _logger.debug("args %s", args)
    env_name: str = args[0]
    context: "FoundationContext" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=FoundationContext)
    _logger.debug("context.name %s", context.name)

    foundation.destroy(context=context)
    _logger.debug("Demo Stack destroyed")
    cdk_toolkit.destroy(context=context)
    _logger.debug("CDK Toolkit Stack destroyed")


def destroy_credentials(args: Tuple[str, ...]) -> None:
    _logger.debug("args: %s", args)
    if len(args) != 2:
        raise ValueError("Unexpected number of values in args")
    env_name: str = args[0]
    registry: str = args[1]
    _logger.debug("Context loaded.")

    secret_id = f"orbit-{env_name}-docker-credentials"
    credentials = secretsmanager.get_secret_value(secret_id=secret_id)

    if registry in credentials:
        del credentials[registry]
        secretsmanager.put_secret_value(secret_id=secret_id, secret=credentials)
        _logger.debug("Registry Credentials destroyed")
    else:
        _logger.debug("Registry Credentials not found, ignoring")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import base64
import logging
import os
import pickle
import shutil
import sys
from typing import TYPE_CHECKING, Any, Dict, List, Type, cast

from aws_orbit import cdk, sh
from aws_orbit.models.context import Context, ContextSerDe, TeamContext
from aws_orbit.services import cfn

if TYPE_CHECKING:
    from aws_cdk.core import Stack

_logger: logging.Logger = logging.getLogger(__name__)


def _serialize_parameters(parameters: Dict[str, Any]) -> str:
    pickled: bytes = pickle.dumps(obj=parameters)
    return base64.b64encode(pickled).decode("utf-8")


def _deserialize_parameters(parameters: str) -> Dict[str, Any]:
    data: bytes = base64.b64decode(parameters.encode("utf-8"))
    return cast(Dict[str, Any], pickle.loads(data))


def cdk_handler(stack_class: Type["Stack"]) -> None:
    _logger.debug("sys.argv: %s", sys.argv)
    if len(sys.argv) != 5:
        raise ValueError(f"Unexpected number of values in sys.argv ({len(sys.argv)}) - {sys.argv}.")

    stack_name: str = sys.argv[1]
    team_name: str = sys.argv[3]
    parameters: Dict[str, Any] = _deserialize_parameters(parameters=sys.argv[4])
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=sys.argv[2], type=Context)
    team_context = context.get_team_by_name(name=team_name)
    if team_context is None:
        raise ValueError(f"Team {team_name} not found in the context.")

    outdir = os.path.join(
        ".orbit.out",
        context.name,
        "cdk",
        stack_name,
    )
    shutil.rmtree(outdir, ignore_errors=True)
    os.makedirs(outdir, exist_ok=True)

    # Can't be imported globally because we only have CDK installed on CodeBuild
    from aws_cdk.core import App

    app = App(outdir=outdir)
    stack_class(app, stack_name, context, team_context, parameters)  # type: ignore
    app.synth(force=True)


def cdk_prep_team_handler(stack_class: Type["Stack"]) -> None:
    _logger.debug("sys.argv: %s", sys.argv)
    if len(sys.argv) != 5:
        raise ValueError(f"Unexpected number of values in sys.argv ({len(sys.argv)}) - {sys.argv}.")

    stack_name: str = sys.argv[1]
    # team_name: str = sys.argv[3]
    parameters: Dict[str, Any] = _deserialize_parameters(parameters=sys.argv[4])
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=sys.argv[2], type=Context)

    # Can not find /orbit/env_name/teams ssm param.
    # team_context = context.get_team_by_name(name=team_name)
    # if team_context is None:
    #     raise ValueError(f"Team {team_name} not found in the context.")

    outdir = os.path.join(
        ".orbit.out",
        context.name,
        "cdk",
        stack_name,
    )
    shutil.rmtree(outdir, ignore_errors=True)
    os.makedirs(outdir, exist_ok=True)

    # Can't be imported globally because we only have CDK installed on CodeBuild
    from aws_cdk.core import App

    app = App(outdir=outdir)
    stack_class(app, stack_name, context, parameters)  # type: ignore
    app.synth(force=True)


def cdk_deploy(
    stack_name: str,
    app_filename: str,
    context: "Context",
    team_context: "TeamContext",
    parameters: Dict[str, Any],
) -> None:
    if context.cdk_toolkit.stack_name is None:
        raise ValueError(f"context.cdk_toolkit_stack_name: {context.cdk_toolkit.stack_name}")
    args: List[str] = [stack_name, context.name, team_context.name, _serialize_parameters(parameters=parameters)]
    cmd: str = (
        "cdk deploy --require-approval never --progress events "
        f"--toolkit-stack-name {context.cdk_toolkit.stack_name} "
        f"{cdk.get_app_argument(app_filename, args)} "
        f"{cdk.get_output_argument(context, stack_name)}"
    )
    sh.run(cmd=cmd)


def cdk_destroy(
    stack_name: str,
    app_filename: str,
    context: "Context",
    team_context: "TeamContext",
    parameters: Dict[str, Any],
) -> None:
    if cfn.does_stack_exist(stack_name=stack_name) is False:
        _logger.debug("Skipping CDK destroy for %s, because the stack was not found.", stack_name)
        return
    if context.cdk_toolkit.stack_name is None:
        raise ValueError(f"context.cdk_toolkit_stack_name: {context.cdk_toolkit.stack_name}")
    args: List[str] = [stack_name, context.name, team_context.name, _serialize_parameters(parameters=parameters)]
    cmd: str = (
        "cdk destroy --force "
        f"--toolkit-stack-name {context.cdk_toolkit.stack_name} "
        f"{cdk.get_app_argument(app_filename, args)} "
        f"{cdk.get_output_argument(context, stack_name)}"
    )
    sh.run(cmd=cmd)



import itertools
import logging
from base64 import b64decode
from typing import Any, Dict, Iterator, List, Optional, Tuple, cast

import boto3

from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def _chunks(iterable: Iterator[Any], size: int) -> Iterator[Any]:
    iterator = iter(iterable)
    for first in iterator:
        yield itertools.chain([first], itertools.islice(iterator, size - 1))


def _filter_repos(env_name: str, page: Dict[str, Any]) -> Iterator[str]:
    client = boto3_client("ecr")
    for repo in page["repositories"]:
        if repo["repositoryName"].startswith(f"orbit-{env_name}/"):
            yield repo["repositoryName"]
        else:
            response: Dict[str, Any] = client.list_tags_for_resource(resourceArn=repo["repositoryArn"])
            for tag in response["tags"]:
                if tag["Key"] == "Env" and tag["Value"] == f"orbit-{env_name}":
                    yield repo["repositoryName"]


def _fetch_repos(env_name: str) -> Iterator[str]:
    client = boto3_client("ecr")
    paginator = client.get_paginator("describe_repositories")
    for page in paginator.paginate():
        for repo_name in _filter_repos(env_name, page=page):
            yield repo_name


def get_credential(region: Optional[str] = None) -> Tuple[str, str]:
    if region is None:
        ecr_client = boto3_client("ecr")
    else:
        _logger.debug("Creating custom boto3 session for region %s", region)
        ecr_client = boto3.Session(region_name=region).client("ecr")
    result = ecr_client.get_authorization_token()
    auth = result["authorizationData"][0]
    auth_token = b64decode(auth["authorizationToken"]).decode()
    return cast(Tuple[str, str], tuple(auth_token.split(sep=":", maxsplit=1)))


def fetch_images(repo: str) -> Iterator[str]:
    client = boto3_client("ecr")
    paginator = client.get_paginator("list_images")
    for page in paginator.paginate(repositoryName=repo):
        for image in page["imageIds"]:
            yield image["imageDigest"]


def delete_images(repo: str) -> None:
    client = boto3_client("ecr")
    for chunk in _chunks(iterable=fetch_images(repo=repo), size=100):
        client.batch_delete_image(repositoryName=repo, imageIds=[{"imageDigest": i} for i in chunk])


def delete_repo(repo: str) -> None:
    client = boto3_client("ecr")
    _logger.debug("Deleting Repository Images: %s", repo)
    delete_images(repo=repo)
    _logger.debug("Deleting Repository: %s", repo)
    client.delete_repository(repositoryName=repo, force=True)


def cleanup_remaining_repos(env_name: str) -> None:
    _logger.debug("Deleting any remaining ECR Repos")
    for repo in _fetch_repos(env_name=env_name):
        delete_repo(repo=repo)


def create_repository(repository_name: str, env_name: Optional[str] = None) -> None:
    client = boto3_client("ecr")
    params: Dict[str, Any] = {"repositoryName": repository_name}
    if env_name:
        params["tags"] = [{"Key": "Env", "Value": env_name}]
    response = client.create_repository(**params)
    if "repository" in response and "repositoryName" in response["repository"]:
        _logger.debug("ECR repository not exist, creating for %s", repository_name)
    else:
        _logger.error("ECR repository creation failed, response %s", response)
        raise RuntimeError(response)


def describe_repositories(repository_names: List[str]) -> List[Dict[str, Any]]:
    client = boto3_client("ecr")
    try:
        return cast(
            List[Dict[str, Any]], client.describe_repositories(repositoryNames=repository_names)["repositories"]
        )
    except client.exceptions.RepositoryNotFoundException:
        return []



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import aws_cdk.aws_cognito as cognito
import aws_cdk.core as core


class CognitoBuilder:
    @staticmethod
    def build_user_pool_group(scope: core.Construct, user_pool_id: str, team_name: str) -> cognito.CfnUserPoolGroup:
        return cognito.CfnUserPoolGroup(
            scope=scope,
            id=f"{team_name}_group",
            user_pool_id=user_pool_id,
            group_name=team_name,
            description=f"{team_name} users group.",
        )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import shutil
from typing import Iterator, List, Optional, cast

import boto3

import aws_orbit
from aws_orbit import ORBIT_CLI_ROOT, cdk, docker, plugins
from aws_orbit.models.context import Context, ContextSerDe, TeamContext, create_team_context_from_manifest
from aws_orbit.models.manifest import Manifest, TeamManifest
from aws_orbit.services import cfn
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def _fetch_targets(context: "Context", fs_id: str) -> Iterator[str]:
    client = boto3_client("efs")
    paginator = client.get_paginator("describe_mount_targets")
    for page in paginator.paginate(FileSystemId=fs_id):
        for target in page["MountTargets"]:
            yield target["MountTargetId"]


def _delete_targets(context: "Context", fs_id: str) -> None:
    client = boto3_client("efs")
    for target in _fetch_targets(context=context, fs_id=fs_id):
        try:
            _logger.debug(f"Deleting MountTargetId: {target}")
            client.delete_mount_target(MountTargetId=target)
        except client.exceptions.MountTargetNotFound:
            _logger.warning(f"Ignoring MountTargetId {target} deletion cause it does not exist anymore.")


def _create_dockerfile(context: "Context", team_context: "TeamContext", image_name: str) -> str:
    if image_name == "jupyter-user":
        base_image_cmd = f"FROM {team_context.base_image_address}"
    else:
        raise Exception(f"The image {image_name} is not deployable to individual Teams.")

    _logger.debug("base_image_cmd: %s", base_image_cmd)
    cmds: List[str] = [base_image_cmd]

    # Add CodeArtifact pip.conf
    cmds += ["USER root"]
    cmds += ["ADD pip.conf /etc/pip.conf"]

    for plugin in team_context.plugins:
        # Adding plugin modules to image via pip
        plugin_module_name = (plugin.module).replace("_", "-")
        cmds += [f"RUN pip install --upgrade aws-orbit-{plugin_module_name}=={aws_orbit.__version__}"]

        hook: plugins.HOOK_TYPE = plugins.PLUGINS_REGISTRIES.get_hook(
            context=context,
            team_name=team_context.name,
            plugin_name=plugin.plugin_id,
            hook_name="dockerfile_injection_hook",
        )
        if hook is not None:
            plugin_cmds = cast(Optional[List[str]], hook(plugin.plugin_id, context, team_context, plugin.parameters))
            if plugin_cmds is not None:
                cmds += [f"# Commands for {plugin.plugin_id} plugin"] + plugin_cmds

    # Removing pip conf and setting to notebook user
    cmds += ["RUN rm /etc/pip.conf", "USER $NB_UID"]
    _logger.debug("Dockerfile cmds: %s", cmds)
    outdir = os.path.join(".orbit.out", context.name, team_context.name, "image")
    output_filename = os.path.join(outdir, "Dockerfile")
    os.makedirs(outdir, exist_ok=True)
    shutil.rmtree(outdir)
    _logger.debug("Writing %s", output_filename)
    os.makedirs(outdir, exist_ok=True)
    content: str = "\n".join(cmds)
    _logger.debug("content:\n%s", content)
    with open(output_filename, "w") as file:
        file.write(content)
    return outdir


def _deploy_team_image(context: "Context", team_context: "TeamContext", image: str) -> None:
    image_dir: str = _create_dockerfile(context=context, team_context=team_context, image_name=image)
    image_name: str = f"orbit-{context.name}-{team_context.name}-{image}"
    _logger.debug("Deploying the %s Docker image", image_name)
    docker.deploy_image_from_source(context=context, dir=image_dir, name=image_name)
    _logger.debug("Docker Image Deployed to ECR (%s).", image_name)


def _deploy_team_bootstrap(context: "Context", team_context: "TeamContext") -> None:
    for plugin in team_context.plugins:
        hook: plugins.HOOK_TYPE = plugins.PLUGINS_REGISTRIES.get_hook(
            context=context,
            team_name=team_context.name,
            plugin_name=plugin.plugin_id,
            hook_name="bootstrap_injection_hook",
        )
        if hook is not None:
            script_content: Optional[str] = cast(
                Optional[str], hook(plugin.plugin_id, context, team_context, plugin.parameters)
            )
            if script_content is not None:
                client = boto3.client("s3")
                key: str = f"{team_context.bootstrap_s3_prefix}{plugin.plugin_id}.sh"
                _logger.debug(f"Uploading s3://{context.toolkit.s3_bucket}/{key}")
                client.put_object(
                    Body=script_content.encode("utf-8"),
                    Bucket=context.toolkit.s3_bucket,
                    Key=key,
                )


def deploy_team(context: "Context", manifest: Manifest, team_manifest: TeamManifest) -> None:
    # Pull team spacific custom cfn plugin, trigger pre_hook
    team_context: Optional["TeamContext"] = create_team_context_from_manifest(
        manifest=manifest, team_manifest=team_manifest
    )
    _logger.debug(f"team_context={team_context}")
    if team_context:
        _logger.debug(f"team_context.plugins={team_context.plugins}")
        _logger.debug("Calling team pre_hook")
        for plugin in team_context.plugins:
            hook: plugins.HOOK_TYPE = plugins.PLUGINS_REGISTRIES.get_hook(
                context=context,
                team_name=team_context.name,
                plugin_name=plugin.plugin_id,
                hook_name="pre_hook",
            )
            if hook is not None:
                _logger.debug(f"Found pre_hook for plugin_id {plugin}")
                hook(plugin.plugin_id, context, team_context, plugin.parameters)
        _logger.debug("End of pre_hook plugin execution")
    else:
        _logger.debug(f"Skipping pre_hook for unknown Team: {team_manifest.name}")

    args = [context.name, team_manifest.name]
    cdk.deploy(
        context=context,
        stack_name=f"orbit-{manifest.name}-{team_manifest.name}",
        app_filename=os.path.join(ORBIT_CLI_ROOT, "remote_files", "cdk", "team.py"),
        args=args,
    )
    team_context = context.get_team_by_name(name=team_manifest.name)
    if team_context:
        team_context.fetch_team_data()
    else:
        team_context = create_team_context_from_manifest(manifest=manifest, team_manifest=team_manifest)
        team_context.fetch_team_data()
        context.teams.append(team_context)

    _logger.debug(
        f"team_context.helm_repository: s3://{context.toolkit.s3_bucket}/helm/repositories/teams/{team_context.name}"
    )
    team_context.team_helm_repository = f"s3://{context.toolkit.s3_bucket}/helm/repositories/teams/{team_context.name}"
    team_context.user_helm_repository = f"s3://{context.toolkit.s3_bucket}/helm/repositories/user/{team_context.name}"
    ContextSerDe.dump_context_to_ssm(context=context)


def destroy_team(context: "Context", team_context: "TeamContext") -> None:
    _logger.debug("Stack name: %s", team_context.stack_name)
    if cfn.does_stack_exist(stack_name=context.toolkit.stack_name):
        if cfn.does_stack_exist(stack_name=team_context.stack_name):
            args: List[str] = [context.name, team_context.name]

            fs_id = cast(str, context.shared_efs_fs_id)
            _delete_efs_endpoints(filesystem_id=fs_id, team_name=team_context.name)

            cdk.destroy(
                context=context,
                stack_name=team_context.stack_name,
                app_filename=os.path.join(ORBIT_CLI_ROOT, "remote_files", "cdk", "team.py"),
                args=args,
            )

        _logger.debug("Team specific post_hook execute to destroy the cfn resources")
        _logger.debug(f"team_context.plugins={team_context.plugins}")
        for plugin in team_context.plugins:
            _logger.debug(f"Checking post hook for plugin={plugin}")
            hook: plugins.HOOK_TYPE = plugins.PLUGINS_REGISTRIES.get_hook(
                context=context,
                team_name=team_context.name,
                plugin_name=plugin.plugin_id,
                hook_name="post_hook",
            )
            if hook is not None:
                _logger.debug(f"Found post hook for team {team_context.name} plugin {plugin.plugin_id}")
                hook(plugin.plugin_id, context, team_context, plugin.parameters)


def destroy_all(context: "Context") -> None:
    for team_context in context.teams:
        destroy_team(context=context, team_context=team_context)
    context.teams = []
    ContextSerDe.dump_context_to_ssm(context=context)


def _delete_efs_endpoints(filesystem_id: str, team_name: str) -> None:
    efs = boto3.client("efs")

    access_points = efs.describe_access_points(FileSystemId=filesystem_id)

    for ap in access_points.get("AccessPoints"):
        tags = ap.get("Tags")
        for tag in tags:
            if tag.get("Key") == "TeamSpace":
                if team_name == tag.get("Value"):
                    access_point_id = ap.get("AccessPointId")
                    _logger.info(f"Deleting EFS endpoint: {access_point_id}")

                    try:
                        efs.delete_access_point(AccessPointId=access_point_id)
                        _logger.info("Access point deleted")
                    except efs.exceptions.AccessPointNotFound:
                        _logger.error("Access point not found")
                    except efs.exceptions.InternalServerError as err:
                        _logger.error(err)
                    except efs.exceptions.BadRequest as err:
                        _logger.error(err)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import Any, Dict, Optional, Sequence, cast

import click
import tqdm

COLOR_ORBIT = "bright_blue"
COLOR_ERROR = "bright_red"
COLOR_WARN = "bright_yellow"

PROGRESS_BAR_FORMAT = "{desc} |{bar}| {percentage:3.0f}% "

_logger: logging.Logger = logging.getLogger(__name__)


def stylize(text: str, color: str = COLOR_ORBIT, bold: bool = False, underline: bool = False) -> str:
    return cast(str, click.style(text=text, bold=bold, underline=underline, fg=color))


def print_list(tittle: str, items: Sequence[str]) -> None:
    click.echo(message=(f"{tittle}\n\n" + "\n".join([f"{stylize('>')} {i}" for i in items])))


REMOTE_PROGRESS_LOOKUP: Dict[str, Dict[str, int]] = {
    "Deploying": {
        "Waiting for agent ping": 20,
        "Waiting for DOWNLOAD_SOURCE": 21,
        "Phase is DOWNLOAD_SOURCE": 22,
        "Phase complete: DOWNLOAD_SOURCE State: SUCCEEDED": 23,
        "Running command npm install -g aws-cdk": 25,
        "Phase complete: INSTALL State: SUCCEEDED": 27,
        "Phase complete: PRE_BUILD State: SUCCEEDED": 30,
        "Manifest loaded": 32,
        "Changeset loaded": 33,
        "Plugins loaded": 34,
        "CDK Toolkit Stack deployed": 40,
        "Demo Stack deployed": 50,
        "Env Stack deployed": 55,
        "Docker images build skipped": 65,
        "Docker Images deployed": 65,
        "EKS Environment Stack deployed": 70,
        "Kubernetes Environment components deployed": 75,
        "Teams Stacks deployed": 80,
        "EKS Team Stacks deployed": 90,
        "Kubernetes Team components deployed": 95,
        "Skipping Team Stacks": 95,
        "Images changeset processed": 96,
        "Phase complete: BUILD State: SUCCEEDED": 97,
    },
    "Destroying": {
        "Waiting for agent ping": 20,
        "Waiting for DOWNLOAD_SOURCE": 21,
        "Phase is DOWNLOAD_SOURCE": 22,
        "Phase complete: DOWNLOAD_SOURCE State: SUCCEEDED": 23,
        "Running command npm install -g aws-cdk": 25,
        "Phase complete: INSTALL State: SUCCEEDED": 27,
        "Phase complete: PRE_BUILD State: SUCCEEDED": 30,
        "Manifest loaded": 31,
        "Changeset loaded": 32,
        "Plugins loaded": 33,
        "Kubernetes Teams components destroyed": 35,
        "EKS Team Stacks destroyed": 40,
        "Teams Stacks destroyed": 50,
        "Kubernetes Environment components destroyed": 60,
        "EKS Environment Stacks destroyed": 70,
        "Env Stack destroyed": 75,
        "Demo Stack destroyed": 85,
        "CDK Toolkit Stack destroyed": 90,
        "Skipping Environment, Demo, and CDK Toolkit Stacks": 90,
        "Phase complete: BUILD State: SUCCEEDED": 94,
    },
    "Deploying Docker Image": {
        "Waiting for agent ping": 20,
        "Waiting for DOWNLOAD_SOURCE": 21,
        "Phase is DOWNLOAD_SOURCE": 22,
        "Phase complete: DOWNLOAD_SOURCE State: SUCCEEDED": 23,
        "Running command npm install -g aws-cdk": 25,
        "Phase complete: INSTALL State: SUCCEEDED": 27,
        "Phase complete: PRE_BUILD State: SUCCEEDED": 30,
        "Plugins loaded": 50,
        "Env changes deployed": 60,
        "Teams Stacks deployed": 75,
        "Phase complete: BUILD State: SUCCEEDED": 99,
    },
    "Destroying Docker Image": {
        "Waiting for agent ping": 20,
        "Waiting for DOWNLOAD_SOURCE": 21,
        "Phase is DOWNLOAD_SOURCE": 22,
        "Phase complete: DOWNLOAD_SOURCE State: SUCCEEDED": 23,
        "Running command npm install -g aws-cdk": 25,
        "Phase complete: INSTALL State: SUCCEEDED": 27,
        "Phase complete: PRE_BUILD State: SUCCEEDED": 30,
        "Env changes deployed": 70,
        "Docker Image Destroyed from ECR": 95,
        "Phase complete: BUILD State: SUCCEEDED": 99,
    },
}


class MessagesContext:
    def __init__(self, task_name: str, debug: bool) -> None:
        self.task_name = task_name
        self.debug = debug
        if self.debug:
            self.pbar = None
        else:
            self.pbar = tqdm.tqdm(
                total=100,
                desc=task_name,
                bar_format=PROGRESS_BAR_FORMAT,
                ncols=50,
                colour="green",
            )

    def __enter__(self) -> "MessagesContext":
        if self.pbar is not None:
            self.pbar.update(1)
        else:
            _logger.debug("Progress bar: 1%")
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        if exc_type is not None:
            self.error(f"{exc_type.__name__}: {exc_value}")
        if self.pbar is not None:
            self.pbar.write("")
            self.pbar.close()
        if exc_traceback is not None:
            click.echo(f"\n{stylize('Error Traceback', color=COLOR_ERROR)}:")

    def info(self, msg: str) -> None:
        tittle: str = stylize(text=" Info ", color="reset", bold=False, underline=False)
        self.echo(tittle=tittle, msg=msg)

    def tip(self, msg: str) -> None:
        tittle: str = stylize(text=" Tip ", color=COLOR_ORBIT, bold=False, underline=False)
        self.echo(tittle=tittle, msg=msg)

    def warn(self, msg: str) -> None:
        tittle: str = stylize(text=" Warn ", color=COLOR_WARN, bold=False, underline=False)
        self.echo(tittle=tittle, msg=msg)

    def error(self, msg: str) -> None:
        tittle: str = stylize(text=" Error ", color=COLOR_ERROR, bold=False, underline=False)
        self.echo(tittle=tittle, msg=msg)

    def echo(self, tittle: str, msg: str) -> None:
        text = f"[{tittle}] {msg}"
        if self.pbar is not None:
            self.pbar.write(text)
        else:
            click.echo(text)

    def progress(self, n: int) -> None:
        if self.pbar is not None:
            current = self.pbar.n
            if current > n:
                raise RuntimeError(f"Current progress: {current} | Desired progress: {n}")
            self.pbar.update(n - current)
        else:
            _logger.debug(f"Progress bar: {n}%")

    def _progress_codebuild_log(self, msg: str) -> bool:
        msg = msg[32:]
        n: Optional[int] = REMOTE_PROGRESS_LOOKUP[self.task_name].get(msg)
        if n is not None:
            self.progress(n=n)
            return True
        return False

    def _progress_cli_log(self, msg: str) -> bool:
        msg_begining = "] "
        if msg_begining in msg:
            msg = msg.split(sep=msg_begining, maxsplit=1)[-1]
            n: Optional[int] = REMOTE_PROGRESS_LOOKUP[self.task_name].get(msg)
            if n is None:
                if msg.startswith("info: "):
                    self.info(msg=msg[6:])
                elif msg.startswith("tip: "):
                    self.tip(msg=msg[5:])
                elif msg.startswith("warn: "):
                    _logger.debug("WARN!")
                    self.warn(msg=msg[6:])
                elif msg.startswith("error: "):
                    self.error(msg=msg[7:])
                else:
                    return False
            else:
                self.progress(n=n)
            return True
        return False

    def progress_bar_callback(self, msg: str) -> None:
        if self._progress_codebuild_log(msg=msg) is False:
            self._progress_cli_log(msg=msg)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
from typing import List, Optional, TypeVar

from aws_orbit import exceptions, sh, utils
from aws_orbit.models.context import Context, FoundationContext
from aws_orbit.services import ecr

_logger: logging.Logger = logging.getLogger(__name__)


T = TypeVar("T")


def login(context: T) -> None:
    if not (isinstance(context, Context) or isinstance(context, FoundationContext)):
        raise ValueError("Unknown 'context' Type")

    username, password = ecr.get_credential()
    ecr_address = f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com"
    sh.run(
        f"docker login --username {username} --password {password} {ecr_address}",
        hide_cmd=True,
    )
    _logger.debug("ECR logged in.")


def login_ecr_only(context: "Context", account_id: Optional[str] = None, region: Optional[str] = None) -> None:
    if account_id is None:
        account_id = context.account_id
    if region is None:
        region = context.region
    username, password = ecr.get_credential(region=region)
    ecr_address = f"{account_id}.dkr.ecr.{region}.amazonaws.com"
    sh.run(
        f"docker login --username {username} --password {password} {ecr_address}",
        hide_cmd=True,
    )
    _logger.debug("ECR logged in (%s / %s).", account_id, region)


def public_pull(name: str, tag: str = "latest") -> None:
    sh.run(f"docker pull {name}:{tag}")


def ecr_pull(context: "Context", name: str, tag: str = "latest") -> None:
    sh.run(f"docker pull {name}:{tag}")


def ecr_pull_external(context: "Context", repository: str, tag: str = "latest") -> None:
    parts: List[str] = repository.split(".")
    if len(parts) < 6:
        raise ValueError(f"Invalid External ECR Repository: {repository}")
    external_account_id: str = parts[0]
    external_region: str = parts[3]
    login_ecr_only(context=context, account_id=external_account_id, region=external_region)
    sh.run(f"docker pull {repository}:{tag}")


def tag_image(context: "Context", remote_name: str, remote_source: str, name: str, tag: str = "latest") -> None:
    ecr_address = f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com"
    if remote_source == "ecr" and not remote_name.startswith("public.ecr.aws"):
        remote_name = f"{ecr_address}/{remote_name}"
    sh.run(f"docker tag {remote_name}:{tag} {ecr_address}/{name}:{tag}")


def build(
    context: "Context",
    dir: str,
    name: str,
    tag: str = "latest",
    use_cache: bool = True,
    pull: bool = False,
    build_args: Optional[List[str]] = None,
) -> None:
    ecr_address = f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com"
    repo_address = f"{ecr_address}/{name}:{tag}"
    cache_str: str = ""
    pull_str: str = "--pull" if pull else ""
    build_args_str = " ".join([f"--build-arg {ba}" for ba in build_args]) if build_args else ""
    if use_cache:
        try:
            ecr_pull(context=context, name=name, tag=tag)
            cache_str = f"--cache-from {repo_address}"
        except exceptions.FailedShellCommand:
            _logger.debug(f"Docker cache not found at ECR {name}:{tag}")
    sh.run(f"docker build {pull_str} {cache_str} {build_args_str} --tag {name}:{tag} .", cwd=dir)


def push(context: "Context", name: str, tag: str = "latest") -> None:
    ecr_address = f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com"
    repo_address = f"{ecr_address}/{name}:{tag}"
    sh.run(f"docker push {repo_address}")


def update_docker_file(context: "Context", dir: str) -> None:
    _logger.debug("Docker directory before building: %s", os.path.abspath(dir))
    utils.print_dir(dir)
    docker_file = os.path.join(dir, "Dockerfile")
    if os.path.exists(docker_file):
        _logger.info("Building DockerFile %s", docker_file)
        tag = context.images.jupyter_user.version
        jupyter_user_base = (
            f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com/orbit-{context.name}/jupyter-user:{tag}"
            if context.images.jupyter_user.get_source(account_id=context.account_id, region=context.region) == "code"
            else f"{context.images.jupyter_user.repository}:{tag}"
        )

        with open(docker_file, "r") as file:
            content: str = file.read()
        content = utils.resolve_parameters(
            content,
            dict(
                region=context.region,
                account=context.account_id,
                env=context.name,
                jupyter_user_base=jupyter_user_base,
            ),
        )
        with open(docker_file, "w") as file:
            file.write(content)


def deploy_image_from_source(
    context: "Context",
    dir: str,
    name: str,
    tag: str = "latest",
    use_cache: bool = True,
    build_args: Optional[List[str]] = None,
) -> None:
    _logger.debug("Adding CodeArtifact login to build environment, used by Dockerfile")
    if not os.path.exists(dir):
        bundle_dir = os.path.join("bundle", dir)
        if os.path.exists(bundle_dir):
            dir = bundle_dir
    if context.codeartifact_domain and context.codeartifact_repository:
        ca_domain: str = context.codeartifact_domain
        ca_repo: str = context.codeartifact_repository
        sh.run(f"aws codeartifact login --tool pip --domain {ca_domain} --repository {ca_repo}")
        sh.run(f"cp ./pip.conf {dir}/")
    build_args = [] if build_args is None else build_args
    _logger.debug("Building docker image from %s", os.path.abspath(dir))
    sh.run(cmd="docker system prune --all --force --volumes")
    update_docker_file(context=context, dir=dir)
    build(context=context, dir=dir, name=name, tag=tag, use_cache=use_cache, pull=True, build_args=build_args)
    _logger.debug("Docker Image built")
    tag_image(context=context, remote_name=name, remote_source="local", name=name, tag=tag)
    _logger.debug("Docker Image tagged")
    push(context=context, name=name, tag=tag)
    _logger.debug("Docker Image pushed")


def replicate_image(
    context: "Context",
    deployed_name: str,
    image_name: str,
    source: Optional[str] = None,
    source_repository: Optional[str] = None,
    source_version: Optional[str] = None,
) -> None:
    _logger.debug("Logged in")
    _logger.debug(f"Context: {vars(context)}")

    attr_name: str = image_name.replace("-", "_")
    if not source:
        final_source = getattr(context.images, attr_name).get_source(
            account_id=context.account_id, region=context.region
        )
    else:
        final_source = source
    if not source_repository:
        final_source_repository = getattr(context.images, attr_name).repository
    else:
        final_source_repository = source_repository
    if not source_version:
        final_source_version = getattr(context.images, attr_name).version
    else:
        final_source_version = source_version

    if final_source == "public":
        public_pull(name=final_source_repository, tag=final_source_version)
        _logger.debug("Pulled Public Image")
    elif final_source in ["ecr", "ecr-internal", "ecr-public"]:
        ecr_pull(context=context, name=final_source_repository, tag=final_source_version)
        _logger.debug("Pulled ECR Image")
    elif final_source == "ecr-external":
        ecr_pull_external(context=context, repository=final_source_repository, tag=final_source_version)
        _logger.debug("Pulled external ECR Image")
    else:
        e = ValueError(
            f"Invalid Image Source: {final_source}. Valid values are: code, public, ecr, ecr-internal, ecr-public"
        )
        _logger.error(e)
        raise e

    tag_image(
        context=context,
        remote_name=final_source_repository,
        remote_source=final_source,
        name=deployed_name,
        tag=final_source_version,
    )
    push(context=context, name=deployed_name, tag=final_source_version)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from enum import Enum
from typing import Callable, Tuple

from aws_orbit.remote_files import build as build_image_module
from aws_orbit.remote_files import delete as delete_image_module
from aws_orbit.remote_files import deploy as deploy_module
from aws_orbit.remote_files import destroy as destroy_module

REMOTE_FUNC_TYPE = Callable[[Tuple[str, ...]], None]


class RemoteCommands(Enum):
    _deploy_image: REMOTE_FUNC_TYPE = deploy_module._deploy_image
    build_image: REMOTE_FUNC_TYPE = build_image_module.build_image
    delete_image: REMOTE_FUNC_TYPE = delete_image_module.delete_image
    deploy_credentials: REMOTE_FUNC_TYPE = deploy_module.deploy_credentials
    deploy_foundation: REMOTE_FUNC_TYPE = deploy_module.deploy_foundation
    deploy_env: REMOTE_FUNC_TYPE = deploy_module.deploy_env
    deploy_teams: REMOTE_FUNC_TYPE = deploy_module.deploy_teams
    destroy_teams: REMOTE_FUNC_TYPE = destroy_module.destroy_teams
    destroy_env: REMOTE_FUNC_TYPE = destroy_module.destroy_env
    destroy_foundation: REMOTE_FUNC_TYPE = destroy_module.destroy_foundation
    destroy_credentials: REMOTE_FUNC_TYPE = destroy_module.destroy_credentials



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import glob
import logging
import os
import shutil
from pprint import pformat
from typing import List, Optional, Tuple

from aws_orbit.models.context import Context

_logger: logging.Logger = logging.getLogger(__name__)


def _is_valid_image_file(file_path: str) -> bool:
    for word in ("/node_modules/", "/build/", "/.mypy_cache/", ".egg-info", "__pycache__"):
        if word in file_path:
            return False
    return True


def _list_files(path: str) -> List[str]:
    path = os.path.join(path, "**")
    return [f for f in glob.iglob(path, recursive=True) if os.path.isfile(f) and _is_valid_image_file(file_path=f)]


def _generate_dir(bundle_dir: str, dir: str, name: str) -> str:
    absolute_dir = os.path.realpath(dir)
    final_dir = os.path.join(bundle_dir, name)
    _logger.debug("absolute_dir: %s", absolute_dir)
    _logger.debug("final_dir: %s", final_dir)
    os.makedirs(final_dir, exist_ok=True)
    shutil.rmtree(final_dir)

    _logger.debug("Copying files to %s", final_dir)
    files: List[str] = _list_files(path=absolute_dir)
    if len(files) == 0:
        raise ValueError(f"{name} ({absolute_dir}) is empty!")
    for file in files:
        _logger.debug(f"***file={file}")
        relpath = os.path.relpath(file, absolute_dir)
        new_file = os.path.join(final_dir, relpath)
        _logger.debug("Copying file to %s", new_file)
        os.makedirs(os.path.dirname(new_file), exist_ok=True)
        _logger.debug("Copying file to %s", new_file)
        shutil.copy(src=file, dst=new_file)

    return final_dir


def generate_bundle(
    command_name: str,
    context: "Context",
    dirs: Optional[List[Tuple[str, str]]] = None,
) -> str:
    remote_dir = os.path.join(os.getcwd(), ".orbit.out", context.name, "remote", command_name)
    bundle_dir = os.path.join(remote_dir, "bundle")
    try:
        shutil.rmtree(bundle_dir)
    except FileNotFoundError:
        pass
    os.makedirs(bundle_dir, exist_ok=True)
    _logger.debug(f"generate_bundle dirs={dirs}")
    # Extra Directories
    if dirs is not None:
        for dir, name in dirs:
            _logger.debug(f"***dir={dir}:name={name}")
            _generate_dir(bundle_dir=bundle_dir, dir=dir, name=name)

    _logger.debug("bundle_dir: %s", bundle_dir)

    files = glob.glob(bundle_dir + "/**", recursive=True)
    _logger.debug("files:\n%s", pformat(files))

    shutil.make_archive(base_name=bundle_dir, format="zip", root_dir=remote_dir, base_dir="bundle")
    return bundle_dir + ".zip"



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import shutil
import time
from typing import Any, Dict, List, Optional, Tuple, cast

import yaml

import aws_orbit
from aws_orbit import ORBIT_CLI_ROOT, exceptions, k8s, sh, utils
from aws_orbit.exceptions import FailedShellCommand
from aws_orbit.models.context import Context, ContextSerDe, TeamContext
from aws_orbit.remote_files import kubeflow
from aws_orbit.remote_files.utils import get_k8s_context
from aws_orbit.services import cfn, elb
from aws_orbit.utils import resolve_parameters

_logger: logging.Logger = logging.getLogger(__name__)


MODELS_PATH = os.path.join(ORBIT_CLI_ROOT, "data", "kubectl")


def _generate_orbit_system_kustomizations(context: "Context", clean_up: bool = True) -> List[str]:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "orbit-system")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    commons_output_path = _generate_orbit_system_commons_manifest(output_path=output_path, context=context)

    return [commons_output_path]


def _orbit_system_commons_base(output_path: str, context: Context) -> None:
    os.makedirs(os.path.join(output_path, "base"), exist_ok=True)
    filenames = ["kustomization.yaml", "00a-commons.yaml", "00b-cert-manager.yaml"]
    for filename in filenames:
        input = os.path.join(MODELS_PATH, "orbit-system", "commons", "base", filename)
        output = os.path.join(output_path, "base", filename)
        with open(input, "r") as file:
            content: str = file.read()
        content = resolve_parameters(
            content,
            dict(
                account_id=context.account_id,
                region=context.region,
                env_name=context.name,
                role_prefix=f"/{context.role_prefix}/" if context.role_prefix else "/",
                secure_port="10260",
            ),
        )
        with open(output, "w") as file:
            file.write(content)


def _generate_orbit_system_commons_manifest(output_path: str, context: "Context") -> str:
    output_path = os.path.join(output_path, "commons")
    os.makedirs(output_path, exist_ok=True)
    _cleanup_output(output_path=output_path)
    _orbit_system_commons_base(output_path=output_path, context=context)
    overlays_path = os.path.join(output_path, "overlays")
    src = (
        os.path.join(MODELS_PATH, "orbit-system", "commons", "overlays", "private")
        if context.networking.data.internet_accessible
        else os.path.join(MODELS_PATH, "orbit-system", "commons", "overlays", "isolated")
    )
    shutil.copytree(
        src=src,
        dst=overlays_path,
    )
    return overlays_path


def _kubeflow_namespaces(context: "Context", clean_up: bool = True) -> str:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "orbit-system")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    filenames = ["kubeflow_namespace.yaml"]

    for filename in filenames:
        input = os.path.join(MODELS_PATH, "kubeflow", filename)
        output = os.path.join(output_path, filename)

        with open(input, "r") as file:
            content: str = file.read()
        content = resolve_parameters(
            content,
            dict(
                env_name=context.name,
                account_id=context.account_id,
                region=context.region,
                sts_ep="legacy" if context.networking.data.internet_accessible else "regional",
            ),
        )
        with open(output, "w") as file:
            file.write(content)

    return output_path


def _orbit_controller(context: "Context", output_path: str) -> None:
    filenames = ["01a-orbit-controller.yaml"]

    for filename in filenames:
        input = os.path.join(MODELS_PATH, "orbit-system", filename)
        output = os.path.join(output_path, filename)

        with open(input, "r") as file:
            content: str = file.read()
        content = resolve_parameters(
            content,
            dict(
                env_name=context.name,
                code_build_image=f"{context.images.code_build.repository}:" f"{context.images.code_build.version}",
                orbit_controller_image=f"{context.images.orbit_controller.repository}:"
                f"{context.images.orbit_controller.version}",
                k8s_utilities_image=f"{context.images.k8s_utilities.repository}:"
                f"{context.images.k8s_utilities.version}",
                image_pull_policy="Always" if aws_orbit.__version__.endswith(".dev0") else "IfNotPresent",
                account_id=context.account_id,
                region=context.region,
                sts_ep="legacy" if context.networking.data.internet_accessible else "regional",
            ),
        )
        with open(output, "w") as file:
            file.write(content)


def _orbit_image_replicator(context: "Context", output_path: str) -> None:
    filenames = ["01b-image-replicator.yaml"]

    for filename in filenames:
        input = os.path.join(MODELS_PATH, "orbit-system", filename)
        output = os.path.join(output_path, filename)

        with open(input, "r") as file:
            content: str = file.read()
        content = resolve_parameters(
            content,
            dict(
                account_id=context.account_id,
                region=context.region,
                env_name=context.name,
                orbit_controller_image=f"{context.images.orbit_controller.repository}:"
                f"{context.images.orbit_controller.version}",
                image_pull_policy="Always" if aws_orbit.__version__.endswith(".dev0") else "IfNotPresent",
            ),
        )
        with open(output, "w") as file:
            file.write(content)


def _ssm_agent_installer(output_path: str, context: "Context") -> None:
    filename = "02a-ssm-agent-daemonset-installer.yaml"
    input = os.path.join(MODELS_PATH, "orbit-system", filename)
    output = os.path.join(output_path, filename)
    shutil.copyfile(src=input, dst=output)


def _sm_operator_installer(output_path: str, context: "Context") -> None:
    filename = "10a-sm-operator.yaml"
    input = os.path.join(MODELS_PATH, "orbit-system", filename)
    output = os.path.join(output_path, filename)
    shutil.copyfile(src=input, dst=output)


def _team(context: "Context", team_context: "TeamContext", output_path: str) -> None:
    input = os.path.join(MODELS_PATH, "teams", "00-team.yaml")
    output = os.path.join(output_path, f"{team_context.name}-00-team.yaml")

    with open(input, "r") as file:
        content: str = file.read()

    content = utils.resolve_parameters(
        content,
        dict(
            team=team_context.name,
            efsid=context.shared_efs_fs_id,
            efsapid=team_context.efs_ap_id,
            efsprivateapid=team_context.efs_private_ap_id if team_context.efs_private_ap_id else "",
            account_id=context.account_id,
            env_name=context.name,
            role_prefix=f"/{context.role_prefix}/" if context.role_prefix else "/",
            team_kms_key_arn=team_context.team_kms_key_arn,
            team_security_group_id=team_context.team_security_group_id,
            cluster_pod_security_group_id=context.cluster_pod_sg_id,
            team_context=ContextSerDe.dump_context_to_str(team_context),
            env_context=ContextSerDe.dump_context_to_str(context),
            region=context.region,
        ),
    )
    _logger.debug("Kubectl Team %s manifest:\n%s", team_context.name, content)
    with open(output, "w") as file:
        file.write(content)

    # team rbac role
    input = os.path.join(MODELS_PATH, "teams", "01-team-rbac-role.yaml")
    output = os.path.join(output_path, f"{team_context.name}-01-team-rbac-role.yaml")

    with open(input, "r") as file:
        content = file.read()
    content = utils.resolve_parameters(content, dict(env_name=context.name, team=team_context.name))
    with open(output, "w") as file:
        file.write(content)

    # bind to admin role
    if team_context.k8_admin:
        # user service account
        input = os.path.join(MODELS_PATH, "teams", "02-admin-binding.yaml")
        output = os.path.join(output_path, f"{team_context.name}-02-admin-binding.yaml")

        with open(input, "r") as file:
            content = file.read()
        content = utils.resolve_parameters(content, dict(team=team_context.name))
        with open(output, "w") as file:
            file.write(content)


def _cleanup_output(output_path: str) -> None:
    files = os.listdir(output_path)
    for file in files:
        if file.endswith(".yaml"):
            os.remove(os.path.join(output_path, file))


def _generate_kube_system_kustomizations(context: "Context", clean_up: bool = True) -> List[str]:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "kube-system")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    efs_output_path = _generate_efs_driver_manifest(output_path=output_path, context=context)
    fsx_output_path = _generate_fsx_driver_manifest(output_path=output_path, context=context)

    return [efs_output_path, fsx_output_path]


def _generate_kube_system_manifest(context: "Context", clean_up: bool = True) -> str:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "kube-system")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    filenames = [
        "00-observability.yaml",
        "01-aws-vgpu-daemonset.yaml",
        "01-nvidia-daemonset.yaml",
        "02-cluster-autoscaler-autodiscover.yaml",
    ]

    for filename in filenames:
        input = os.path.join(MODELS_PATH, "kube-system", filename)
        output = os.path.join(output_path, filename)

        with open(input, "r") as file:
            content: str = file.read()
        content = utils.resolve_parameters(
            content,
            dict(
                account_id=context.account_id,
                region=context.region,
                role_prefix=f"/{context.role_prefix}/" if context.role_prefix else "/",
                env_name=context.name,
                cluster_name=f"orbit-{context.name}",
                sts_ep="legacy" if context.networking.data.internet_accessible else "regional",
                image_pull_policy="Always" if aws_orbit.__version__.endswith(".dev0") else "IfNotPresent",
                use_static_instance_list=str(not context.networking.data.internet_accessible).lower(),
            ),
        )
        with open(output, "w") as file:
            file.write(content)

    return output_path


def set_secondary_vpc_cidr_env_vars() -> None:
    sh.run("kubectl set env daemonset aws-node -n kube-system AWS_VPC_K8S_CNI_CUSTOM_NETWORK_CFG=true")
    sh.run("kubectl set env daemonset aws-node -n kube-system ENI_CONFIG_LABEL_DEF=topology.kubernetes.io/zone")


def deploy_eniconfig(az_subnet_map: Dict[str, str], context: "Context") -> None:
    _logger.debug("Prepping eniconfig deployments...")
    output_path = _generate_eni_configs(az_subnet_map, context, clean_up=True)
    sh.run(f"kubectl apply -f {output_path}")
    _logger.debug("Deployted eniconfig deployments...")


def _generate_eni_configs(az_subnet_map: Dict[str, str], context: "Context", clean_up: bool) -> str:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "networking")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    output = os.path.join(output_path, "eniconfigs.yaml")
    for subnet, az in az_subnet_map.items():
        config = {
            "apiVersion": "crd.k8s.amazonaws.com/v1alpha1",
            "kind": "ENIConfig",
            "metadata": {"name": az},
            "spec": {"securityGroups": [context.cluster_pod_sg_id], "subnet": subnet},
        }
        _logger.debug(f"Eniconfig deployments adding...{config}")
        with open(output, "a") as outfile:
            yaml.dump(config, outfile, default_flow_style=False)
            outfile.write("---")
            outfile.write("\n")
    return output


def _generate_orbit_system_manifest(context: "Context", clean_up: bool = True) -> str:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "orbit-system")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    _orbit_controller(context=context, output_path=output_path)

    if context.install_ssm_agent:
        _logger.debug("Deploying SSM Agent Installer")
        _ssm_agent_installer(output_path=output_path, context=context)
    else:
        _logger.debug("Skipping deployment of SSM Agent Installer")

    _sm_operator_installer(output_path=output_path, context=context)

    return output_path


def _generate_orbit_image_replicator_manifest(context: "Context", clean_up: bool = True) -> Optional[str]:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "orbit-system")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    if context.install_image_replicator or not context.networking.data.internet_accessible:
        _logger.debug("Deploying Pod Image Modifier and Image Replicator")
        _orbit_image_replicator(output_path=output_path, context=context)
        return output_path
    else:
        _logger.debug("Skipping deployment of Pod Image Modifier and Image Replicator")
        return None


def _generate_orbit_system_env_kustomizations(context: "Context", clean_up: bool = True) -> List[str]:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "env")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    env_output_path = _generate_orbit_system_env_manifest(output_path=output_path, context=context)

    return [env_output_path]


def _orbit_system_env_base(output_path: str, context: Context) -> None:
    os.makedirs(os.path.join(output_path, "base"), exist_ok=True)
    filenames = ["kustomization.yaml", "00-commons.yaml"]
    for filename in filenames:
        input = os.path.join(MODELS_PATH, "env", "base", filename)
        output = os.path.join(output_path, "base", filename)
        with open(input, "r") as file:
            content: str = file.read()
        content = utils.resolve_parameters(
            content,
            dict(
                env_name=context.name,
                orbit_controller_image=f"{context.images.orbit_controller.repository}:"
                f"{context.images.orbit_controller.version}",
                k8s_utilities_image=f"{context.images.k8s_utilities.repository}:"
                f"{context.images.k8s_utilities.version}",
                image_pull_policy="Always" if aws_orbit.__version__.endswith(".dev0") else "IfNotPresent",
                certArn=context.networking.frontend.ssl_cert_arn,
                cognitoAppClientId=context.user_pool_client_id,
                cognitoUserPoolID=context.user_pool_id,
                account_id=context.account_id,
                region=context.region,
                cognitoUserPoolDomain=context.cognito_external_provider_domain,
            ),
        )
        with open(output, "w") as file:
            file.write(content)


def _generate_orbit_system_env_manifest(output_path: str, context: "Context") -> str:
    _cleanup_output(output_path=output_path)
    _orbit_system_env_base(output_path=output_path, context=context)
    overlays_path = os.path.join(output_path, "overlays")
    src = (
        os.path.join(MODELS_PATH, "env", "overlays", "private")
        if context.networking.data.internet_accessible
        else os.path.join(MODELS_PATH, "env", "overlays", "isolated")
    )
    shutil.copytree(
        src=src,
        dst=overlays_path,
    )
    return overlays_path


def _generate_kubeflow_patch(context: "Context", clean_up: bool = True) -> Tuple[str, str]:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "env")
    os.makedirs(output_path, exist_ok=True)
    if clean_up:
        _cleanup_output(output_path=output_path)

    # kubeflow jupyter launcher configmap
    input = os.path.join(MODELS_PATH, "kubeflow", "kf-jupyter-launcher.yaml")
    output = os.path.join(output_path, "kf-jupyter-launcher.yaml")
    with open(input, "r") as file:
        content = file.read()

    content = utils.resolve_parameters(
        content,
        dict(
            orbit_jupyter_user_image=f"{context.images.jupyter_user.repository}:{context.images.jupyter_user.version}"
        ),
    )
    with open(output, "w") as file:
        file.write(content)

    input = os.path.join(MODELS_PATH, "kubeflow", "kf-jupyter-patch.yaml")

    with open(input, "r") as file:
        patch = file.read()

    return output, patch


def _prepare_team_context_path(context: "Context") -> str:
    output_path = os.path.join(".orbit.out", context.name, "kubectl", "apps")
    os.makedirs(output_path, exist_ok=True)
    _cleanup_output(output_path=output_path)
    if context.account_id is None:
        raise ValueError("context.account_id is None!")
    return output_path


def _generate_teams_manifest(context: "Context") -> str:
    output_path: str = _prepare_team_context_path(context=context)
    for team_context in context.teams:
        _team(context=context, team_context=team_context, output_path=output_path)
    return output_path


def _generate_team_context(context: "Context", team_context: "TeamContext") -> str:
    output_path: str = _prepare_team_context_path(context=context)
    _team(context=context, team_context=team_context, output_path=output_path)
    return output_path


def _update_elbs(context: "Context") -> None:
    elbs: Dict[str, Dict[str, Any]] = elb.get_elbs_by_service(env_name=context.name)
    # Env ELBs
    context.elbs = {k: v for k, v in elbs.items() if k.startswith("env/")}
    # Teams ELBS
    for team in context.teams:
        team.elbs = {k: v for k, v in elbs.items() if k.startswith(f"{team.name}/")}


def fetch_kubectl_data(context: "Context", k8s_context: str) -> None:
    _logger.debug("Fetching Kubectl data...")

    ingress_url: str = k8s.get_ingress_dns(name="istio-ingress", k8s_context=k8s_context, namespace="istio-system")

    if context.networking.frontend.custom_domain_name:
        context.landing_page_url = f"https://{context.networking.frontend.custom_domain_name}"
    else:
        context.landing_page_url = f"https://{ingress_url}"

    if context.cognito_external_provider:
        context.cognito_external_provider_redirect = context.landing_page_url

    _update_elbs(context=context)

    ContextSerDe.dump_context_to_ssm(context=context)
    _logger.debug("Kubectl data fetched successfully.")


def _efs_driver_base(output_path: str) -> None:
    os.makedirs(os.path.join(output_path, "base"), exist_ok=True)
    filenames = ("csidriver.yaml", "kustomization.yaml", "node.yaml")
    for filename in filenames:
        input = os.path.join(MODELS_PATH, "kube-system", "efs_driver", "base", filename)
        output = os.path.join(output_path, "base", filename)
        _logger.debug("Copying efs driver base file: %s -> %s", input, output)
        shutil.copyfile(src=input, dst=output)


def _generate_efs_driver_manifest(output_path: str, context: "Context") -> str:
    output_path = os.path.join(output_path, "efs_driver")
    os.makedirs(output_path, exist_ok=True)
    _cleanup_output(output_path=output_path)
    if context.account_id is None:
        raise RuntimeError("context.account_id is None!")
    if context.region is None:
        raise RuntimeError("context.region is None!")
    _efs_driver_base(output_path=output_path)
    overlays_path = os.path.join(output_path, "overlays")
    os.makedirs(overlays_path, exist_ok=True)
    shutil.copyfile(
        src=os.path.join(MODELS_PATH, "kube-system", "efs_driver", "overlays", "kustomization.yaml"),
        dst=os.path.join(overlays_path, "kustomization.yaml"),
    )
    return overlays_path


def _fsx_driver_base(output_path: str, context: "Context") -> None:
    os.makedirs(os.path.join(output_path, "base"), exist_ok=True)
    filenames = [
        "controller-deployment.yaml",
        "controller-serviceaccount.yaml",
        "csidriver.yaml",
        "kustomization.yaml",
        "node-daemonset.yaml",
        "node-serviceaccount.yaml",
    ]
    for filename in filenames:
        input = os.path.join(MODELS_PATH, "kube-system", "fsx_driver", "base", filename)
        output = os.path.join(output_path, "base", filename)
        _logger.debug("Copying fsx driver base file: %s -> %s", input, output)
        with open(input, "r") as file:
            content: str = file.read()
        content = utils.resolve_parameters(
            content,
            dict(
                orbit_cluster_role=context.eks_cluster_role_arn,
            ),
        )
        with open(output, "w") as file:
            file.write(content)


def _generate_fsx_driver_manifest(output_path: str, context: "Context") -> str:
    output_path = os.path.join(output_path, "fsx_driver")
    os.makedirs(output_path, exist_ok=True)
    _cleanup_output(output_path=output_path)
    if context.account_id is None:
        raise RuntimeError("context.account_id is None!")
    if context.region is None:
        raise RuntimeError("context.region is None!")
    _fsx_driver_base(output_path=output_path, context=context)
    overlays_path = os.path.join(output_path, "overlays")
    os.makedirs(overlays_path, exist_ok=True)
    shutil.copyfile(
        src=os.path.join(MODELS_PATH, "kube-system", "fsx_driver", "overlays", "stable", "kustomization.yaml"),
        dst=os.path.join(overlays_path, "kustomization.yaml"),
    )
    return overlays_path


# Confirm orbit-system Service Endpoints
def _confirm_endpoints(name: str, namespace: str, k8s_context: str) -> None:
    addresses = k8s.get_service_addresses(name=name, namespace=namespace, k8s_context=k8s_context)
    if addresses:
        for address in addresses:
            _logger.debug(
                "Service: %s Namespace: %s Hostname: %s IP: %s",
                name,
                namespace,
                address.get("hostname"),
                address.get("ip"),
            )
    else:
        raise Exception("No Endpoints found for Service: %s Namespace: %s", name, namespace)


def _confirm_readiness(name: str, namespace: str, type: str, k8s_context: str) -> None:
    for _ in range(20):
        status = k8s.get_resource_status(name=name, namespace=namespace, type=type, k8s_context=k8s_context)
        ready_replicas = status.get("ready_replicas")
        if ready_replicas and int(ready_replicas) > 0:
            _logger.debug("%s/%s ready", namespace, name)
            break
        else:
            _logger.debug("%s/%s not yet ready, sleeping for 1 minute", namespace, name)
            time.sleep(60)
    else:
        raise Exception("Timeout wating for Image Replicator to become ready")


def write_kubeconfig(context: "Context") -> None:
    sh.run(f"eksctl utils write-kubeconfig --cluster orbit-{context.name} --set-kubeconfig-context")


def deploy_env(context: "Context") -> None:
    eks_stack_name: str = f"eksctl-orbit-{context.name}-cluster"
    _logger.debug("EKSCTL stack name: %s", eks_stack_name)
    if cfn.does_stack_exist(stack_name=eks_stack_name):
        k8s_context = get_k8s_context(context=context)
        _logger.debug("k8s_context: %s", k8s_context)

        # orbit-system kustomizations
        output_paths = _generate_orbit_system_kustomizations(context=context)
        for path in output_paths:
            sh.run(f"kubectl apply -k {path} --context {k8s_context} --wait")

        # Wait until cert-manager webhook is available
        _confirm_endpoints(name="cert-manager-webhook", namespace="cert-manager", k8s_context=k8s_context)
        _confirm_readiness(name="cert-manager", namespace="cert-manager", type="Deployment", k8s_context=k8s_context)
        _confirm_readiness(
            name="cert-manager-cainjector", namespace="cert-manager", type="Deployment", k8s_context=k8s_context
        )

        output_path: Optional[str] = _generate_orbit_system_manifest(context=context, clean_up=True)
        sh.run(f"kubectl apply -f {output_path} --context {k8s_context} --wait")

        output_path = _generate_orbit_image_replicator_manifest(context=context, clean_up=True)
        if output_path is not None:
            sh.run(f"kubectl apply -f {output_path} --context {k8s_context} --wait")

        # Commented until we confirm this isn't needed
        # Restart orbit-system deployments and statefulsets to force reload of caches etc
        # sh.run(f"kubectl rollout restart deployments -n orbit-system --context {k8s_context}")

        _confirm_readiness(
            name="podsetting-operator", namespace="orbit-system", type="deployment", k8s_context=k8s_context
        )
        _confirm_readiness(
            name="teamspace-operator", namespace="orbit-system", type="deployment", k8s_context=k8s_context
        )
        _confirm_readiness(
            name="userspace-operator", namespace="orbit-system", type="deployment", k8s_context=k8s_context
        )
        _confirm_endpoints(name="podsetting-pod-webhook", namespace="orbit-system", k8s_context=k8s_context)

        if context.install_image_replicator or not context.networking.data.internet_accessible:
            _confirm_readiness(
                name="imagereplication-operator", namespace="orbit-system", type="deployment", k8s_context=k8s_context
            )
            _confirm_endpoints(name="imagereplication-pod-webhook", namespace="orbit-system", k8s_context=k8s_context)
            sh.run(
                "kubectl rollout restart daemonsets -n orbit-system-ssm-daemons "
                f"ssm-agent-installer --context {k8s_context}"
            )

        # kube-system kustomizations
        output_paths = _generate_kube_system_kustomizations(context=context)
        for output_path in output_paths:
            sh.run(f"kubectl apply -k {output_path} --context {k8s_context} --wait")

        # kube-system manifests
        output_path = _generate_kube_system_manifest(context=context)
        sh.run(f"kubectl apply -f {output_path} --context {k8s_context} --wait")

        # Enable ENIs
        _enable_eni(k8s_context=k8s_context)

        # kubeflow-namespaces
        output_path = _kubeflow_namespaces(context=context)
        sh.run(f"kubectl apply -f {output_path} --context {k8s_context} --wait")

        kubeflow.deploy_kubeflow(context=context)

        # env
        output_paths = _generate_orbit_system_env_kustomizations(context=context)
        for output_path in output_paths:
            sh.run(f"kubectl apply -k {output_path} --context {k8s_context} --wait")

        # Patch Kubeflow
        _logger.debug("Orbit applying KubeFlow patch")
        jupyter_launcher_config_map, patch = _generate_kubeflow_patch(context=context)
        sh.run(f"kubectl apply -f {jupyter_launcher_config_map} --context {k8s_context} --wait")
        sh.run(f'kubectl patch deployment -n kubeflow jupyter-web-app-deployment --patch "{patch}"')
        sh.run("kubectl rollout restart deployment jupyter-web-app-deployment -n kubeflow")

        _apply_deployment_patch_force_env_nodes("istio-system")
        _apply_deployment_patch_force_env_nodes("knative-serving")
        _apply_deployment_patch_force_env_nodes("kube-system")
        _apply_deployment_patch_force_env_nodes("kubeflow")

        # Patch Pods to push into Fargate when deploying in an isolated subnet
        if not context.networking.data.internet_accessible:
            patch = (
                '{"spec":{"template":{"metadata":{"labels":{"orbit/node-type":"fargate"}},'
                '"spec":{"nodeSelector": null}}}}'
            )
            sh.run(f"kubectl patch deployment -n istio-system authzadaptor --patch '{patch}'")

            patch = (
                '{"spec":{"template":{"metadata":{"labels":{"orbit/node-type":"fargate"}},'
                '"spec":{"nodeSelector": null, "containers":[{"name":"alb-ingress-controller","args":'
                '["--ingress-class=alb","--cluster-name=$(CLUSTER_NAME)","--aws-vpc-id=VPC_ID"]}]}}}}'
            )
            patch = patch.replace("VPC_ID", cast(str, context.networking.vpc_id))
            sh.run(f"kubectl patch deployment -n kubeflow alb-ingress-controller --patch '{patch}'")

        # Patch the kubeflow mpi-operator deployment to version lock the images to v0.2.3
        patch = (
            '{"spec":{"template":{"spec":{"containers":[{"name":"mpi-operator","args":["-alsologtostderr",'
            '"--lock-namespace","kubeflow","--kubectl-delivery-image","mpioperator/kubectl-delivery:v0.2.3"],'
            '"image":"mpioperator/mpi-operator:v0.2.3"}]}}}}'
        )
        sh.run(f"kubectl patch deployment -n kubeflow mpi-operator --patch '{patch}'")

        # Confirm env Service Endpoints
        _confirm_endpoints(name="landing-page-service", namespace="orbit-system", k8s_context=k8s_context)


def _enable_eni(k8s_context: str) -> None:
    _logger.debug("Setting aws-node daemonset in kube-system -- ENABLE_POD_ENI=true")
    sh.run(f"kubectl set env daemonset aws-node -n kube-system --context {k8s_context} ENABLE_POD_ENI=true")

    _logger.debug("Patch aws-node daemonset, container aws-vpc-cni-init -- DISABLE_TCP_EARLY_DEMUX=true")
    patch = (
        '{"spec": {"template": {"spec": {"initContainers": '
        '[{"env":[{"name":"DISABLE_TCP_EARLY_DEMUX","value":"true"}],"name":"aws-vpc-cni-init"}]}}}}'
    )
    sh.run(f"kubectl patch daemonset aws-node -n kube-system --context {k8s_context} --patch '{patch}'")


def _apply_deployment_patch_force_env_nodes(namespace: str) -> None:
    _logger.debug(f"Force {namespace} deployments to env nodes")
    patch = (
        '{"spec":{"template":{"metadata":{"labels":{"orbit/node-type":"ec2"}},'
        '"spec":{"nodeSelector":{"orbit/usage":"reserved","orbit/node-group":"env"}}}}}'
    )
    try:
        sh.run(
            f"bash -c 'for dep in $(kubectl get deployments -n {namespace} "
            f"--output=jsonpath={{.items..metadata.name}}); "
            f"do kubectl patch deployment $dep -n {namespace} --patch '\"'\"'{patch}'\"'\"'; "
            f"done'"
        )

    except FailedShellCommand:
        _logger.debug(f"Ignoring failed moving of {namespace} pods to env nodes")


def deploy_team(context: "Context", team_context: "TeamContext") -> None:
    eks_stack_name: str = f"eksctl-orbit-{context.name}-cluster"
    _logger.debug("EKSCTL stack name: %s", eks_stack_name)
    if cfn.does_stack_exist(stack_name=eks_stack_name):
        k8s_context = get_k8s_context(context=context)
        _logger.debug("kubectl context: %s", k8s_context)
        output_path = _generate_team_context(context=context, team_context=team_context)
        sh.run(f"kubectl apply -f {output_path} --context {k8s_context} --wait")


def destroy_env(context: "Context") -> None:
    eks_stack_name: str = f"eksctl-orbit-{context.name}-cluster"
    _logger.debug("EKSCTL stack name: %s", eks_stack_name)
    if cfn.does_stack_exist(stack_name=eks_stack_name):
        sh.run(f"eksctl utils write-kubeconfig --cluster orbit-{context.name} --set-kubeconfig-context")
        k8s_context = get_k8s_context(context=context)
        _logger.debug("kubectl k8s_context: %s", k8s_context)
        try:
            # Here we remove some finalizers that can cause our delete to hang indefinitely
            try:
                sh.run(
                    "kubectl patch crd/trainingjobs.sagemaker.aws.amazon.com "
                    '--patch \'{"metadata":{"finalizers":[]}}\' --type=merge'
                    f" --context {k8s_context}"
                )
            except FailedShellCommand:
                _logger.debug("Ignoring patch failure")

            output_path = _generate_orbit_system_manifest(context=context)
            sh.run(
                f"kubectl delete -f {output_path} --grace-period=0 --force "
                f"--ignore-not-found --wait --context {k8s_context}"
            )
            output_paths = _generate_orbit_system_kustomizations(context=context, clean_up=True)
            for output_path in output_paths:
                sh.run(
                    f"kubectl delete -k {output_path} --grace-period=0 --force "
                    f"--ignore-not-found --wait --context {k8s_context}"
                )

        except exceptions.FailedShellCommand as ex:
            _logger.debug("Skipping: %s", ex)
            pass  # Let's leave for eksctl, it will destroy everything anyway...


def destroy_teams(context: "Context") -> None:
    eks_stack_name: str = f"eksctl-orbit-{context.name}-cluster"
    _logger.debug("EKSCTL stack name: %s", eks_stack_name)
    if cfn.does_stack_exist(stack_name=eks_stack_name):
        sh.run(f"eksctl utils write-kubeconfig --cluster orbit-{context.name} --set-kubeconfig-context")
        k8s_context = get_k8s_context(context=context)
        _logger.debug("kubectl k8s_context: %s", k8s_context)
        _logger.debug("Attempting kubectl delete")
        output_path = _generate_teams_manifest(context=context)
        utils.print_dir(dir=output_path)
        try:
            sh.run(
                f"kubectl delete -f {output_path} --grace-period=0 --force "
                f"--ignore-not-found --wait=false --context {k8s_context}"
            )
        except exceptions.FailedShellCommand as ex:
            _logger.debug("Skipping: %s", ex)
            pass  # Let's leave for eksctl, it will destroy everything anyway...


def destroy_team(context: "Context", team_context: "TeamContext") -> None:
    eks_stack_name: str = f"eksctl-orbit-{context.name}-cluster"
    _logger.debug("EKSCTL stack name: %s", eks_stack_name)
    if cfn.does_stack_exist(stack_name=eks_stack_name):
        k8s_context = get_k8s_context(context=context)
        _logger.debug("kubectl k8s_context: %s", k8s_context)
        _logger.debug("Attempting kubectl delete for team %s", team_context.name)
        output_path = _generate_team_context(context=context, team_context=team_context)
        sh.run(
            f"kubectl delete -f {output_path} --grace-period=0 --force "
            f"--ignore-not-found --wait --context {k8s_context}"
        )
        # Destory all related user spaces
        sh.run(f"kubectl delete namespaces -l orbit/team={team_context.name} --context {k8s_context}")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import TYPE_CHECKING, cast

import aws_cdk.core as core
from aws_cdk import aws_lambda

from aws_orbit.remote_files.cdk import _lambda_path

if TYPE_CHECKING:
    from aws_orbit.models.context import Context


class LambdaBuilder:
    @staticmethod
    def get_or_build_construct_request(
        scope: core.Construct,
        context: "Context",
        team_name: str,
    ) -> aws_lambda.Function:
        stack = core.Stack.of(cast(core.IConstruct, scope))
        lambda_function = cast(aws_lambda.Function, stack.node.try_find_child("construct_request"))
        if lambda_function is None:
            lambda_function = aws_lambda.Function(
                scope=stack,
                id="construct_request",
                function_name=f"orbit-{context.name}-{team_name}-k8s-construct-request",
                code=aws_lambda.Code.asset(_lambda_path("construct_request")),
                handler="index.handler",
                runtime=aws_lambda.Runtime.PYTHON_3_6,
                timeout=core.Duration.seconds(10),
            )
        return lambda_function



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import Any, Dict, Optional, cast

from aws_orbit.models.manifest import ManagedNodeGroupManifest
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def get_nodegroup_autoscaling_group(cluster_name: str, nodegroup_name: str) -> Optional[Dict[str, Any]]:
    _logger.debug(f"Getting AutoScaling Group for Cluster: {cluster_name}, NodeGroup: {nodegroup_name}")

    client = boto3_client("autoscaling")
    paginator = client.get_paginator("describe_auto_scaling_groups")

    for response in paginator.paginate():
        for asg in response["AutoScalingGroups"]:
            found_cluster_tag = False
            found_nodegroup_tag = False
            for tag in asg["Tags"]:
                if tag["Key"] == "eks:cluster-name" and tag["Value"] == cluster_name:
                    found_cluster_tag = True
                if tag["Key"] == "eks:nodegroup-name" and tag["Value"] == nodegroup_name:
                    found_nodegroup_tag = True

            if found_cluster_tag and found_nodegroup_tag:
                _logger.debug(f"Found AutoScaling Group: {asg['AutoScalingGroupName']}")
                return cast(Dict[str, Any], asg)

    return None


def update_nodegroup_autoscaling_group(cluster_name: str, nodegroup_manifest: ManagedNodeGroupManifest) -> None:
    _logger.debug(f"Updating AutoScaling Group for Cluster: {cluster_name}, NodeGroup: {nodegroup_manifest.name}")
    _logger.debug(
        f"DesiredCapacity: {nodegroup_manifest.nodes_num_desired}, Min: {nodegroup_manifest.nodes_num_min}, "
        f"Max: {nodegroup_manifest.nodes_num_max}"
    )
    asg = get_nodegroup_autoscaling_group(cluster_name=cluster_name, nodegroup_name=nodegroup_manifest.name)

    if not asg:
        _logger.debug(f"No AutoScaling Group found for Cluster: {cluster_name}, NodeGroup: {nodegroup_manifest.name}")
        return

    client = boto3_client("autoscaling")
    client.update_auto_scaling_group(
        AutoScalingGroupName=asg["AutoScalingGroupName"],
        MinSize=nodegroup_manifest.nodes_num_min,
        MaxSize=nodegroup_manifest.nodes_num_max,
        DesiredCapacity=nodegroup_manifest.nodes_num_desired,
    )



""" Module to deploy self signed certs and upload it to IAM """
import logging
import subprocess
from typing import cast

import boto3
from botocore.exceptions import ClientError

from aws_orbit.models.context import FoundationContext

_logger: logging.Logger = logging.getLogger(__name__)


def run_command(cmd: str) -> str:
    """Module to run shell commands."""
    try:
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True, timeout=3, universal_newlines=True)
    except subprocess.CalledProcessError as exc:
        _logger.debug("Command failed with exit code {}, stderr: {}".format(exc.returncode, exc.output.decode("utf-8")))
        raise Exception(exc.output.decode("utf-8"))
    return output


def check_cert(context: "FoundationContext") -> str:
    ssl_cert_arn: str = cast(str, context.networking.frontend.ssl_cert_arn)
    if ssl_cert_arn:
        return ssl_cert_arn
    else:
        return deploy_selfsigned_cert(context)


def deploy_selfsigned_cert(context: "FoundationContext") -> str:
    """Module to deploy self signed cert"""

    _logger.debug("Generating self-signed certificate...")

    _logger.debug("Creating private key...")
    convert_key: str = "openssl genrsa 2048 > private.pem"
    run_command(cmd=convert_key)

    _logger.debug("Creating Certificate signing request...")
    generate_csr: str = (
        "openssl req -new -key private.pem "
        "-out csr.pem "
        '-subj "/C=US/ST=SEA/L=SEA/O=AWSOrbit Security/OU=AWSOrbit Department/CN=AWSOrbit"'
    )
    run_command(cmd=generate_csr)

    _logger.debug("Creating Public certificate...")
    generate_pub: str = "openssl x509 -req -days 365 -in csr.pem " "-signkey private.pem -out public.pem "
    run_command(cmd=generate_pub)

    with open("private.pem", "r") as fp:
        private_pem = fp.read()

    with open("public.pem", "r") as fp:
        public_pem = fp.read()

    _logger.debug("Uploading the cert to IAM...")
    ssl_cert_arn = upload_cert_iam(context, private_pem, public_pem)
    return ssl_cert_arn


def upload_cert_iam(context: "FoundationContext", private_pem: str, public_pem: str) -> str:
    """Uploads the cert to AWS IAM"""
    iam_client = boto3.client("iam")
    ssl_cert_name = f"{context.name}-{context.region}"
    try:
        response = iam_client.get_server_certificate(ServerCertificateName=ssl_cert_name)
        return cast(str, response["ServerCertificate"]["ServerCertificateMetadata"]["Arn"])
    except ClientError as ex:
        if ex.response["Error"]["Code"] == "NoSuchEntity":
            response = iam_client.upload_server_certificate(
                ServerCertificateName=ssl_cert_name, CertificateBody=public_pem, PrivateKey=private_pem
            )
            return cast(str, response["ServerCertificateMetadata"]["Arn"])
        else:
            raise ex
